import json
import logging
import traceback
from pathlib import Path

import hydra
from hydra.utils import instantiate
from omegaconf import DictConfig
from tqdm import tqdm

from tapeagents.io import load_tapes, stream_yaml_tapes
from tapeagents.parallel_processing import lazy_thread_pool_processor

from ..tape import FormFillerAgentMetadata, FormFillerTape, FormFillerUserMetadata
from ..user_simulator_agent import (
    SampleUserInstructionThought,
    UserSimulatorAgent,
    UserSimulatorError,
    UserSimulatorTape,
)

logger = logging.getLogger(__name__)


def run_user_simulator_agent(
    form_filler_tape: FormFillerTape, user_simulator_agent: UserSimulatorAgent
) -> tuple[None | Exception, FormFillerTape, None | UserSimulatorTape]:
    # assert this tape is a user tape or user_simulator_agent tape
    assert form_filler_tape.context is not None, "Tape context is None"
    assert isinstance(
        form_filler_tape.metadata, (FormFillerAgentMetadata)
    ), "Tape metadata is not of type FormFillerUserMetadata or FormFillerAgentMetadata"

    user_simulator_agent_tape = None
    try:
        # Convert formfiller tape to UserSimulatorAgent tape
        user_simulator_agent_tape = user_simulator_agent.make_own_tape(form_filler_tape)

        user_simulator_agent_tape = user_simulator_agent.run(user_simulator_agent_tape).get_final_tape()

        # Reconvert to formfiller tape
        continued_tape = user_simulator_agent.add_observation(form_filler_tape, user_simulator_agent_tape)
        # fix the metadata to be the correct type
        continued_tape.metadata = FormFillerUserMetadata(
            **continued_tape.metadata.model_dump(),
            last_agent_tape_id=form_filler_tape.metadata.id,
        )
        # recover the primary and secondary user behaviors used
        continued_tape.metadata.user_behavior = user_simulator_agent.behavior_alias
        if isinstance(sample_step := user_simulator_agent_tape.steps[-2], SampleUserInstructionThought):
            continued_tape.metadata.user_secondary_behavior = sample_step.instruction_alias

    except Exception as e:
        # Return input formfiller tape
        if user_simulator_agent_tape is not None:  # if make_own_tape didn't fail
            user_simulator_agent_tape.steps.append(UserSimulatorError(error=str(e) + "\n" + traceback.format_exc()))

        return e, form_filler_tape, user_simulator_agent_tape
    return None, continued_tape, user_simulator_agent_tape


@hydra.main(config_path="../conf", config_name="run_user_simulator", version_base="1.2")
def main(cfg: DictConfig):
    output_path = Path(cfg.output_path)  # must be a directory
    output_path.mkdir(exist_ok=True, parents=True)

    logger.info(f"Hydra runtime dir: {hydra.core.hydra_config.HydraConfig.get().runtime.output_dir}")
    logger.info(f"Output path: {output_path}")

    input_dialogues_path = cfg.input_dialogues_path

    tapes: list[FormFillerTape] = load_tapes(FormFillerTape, input_dialogues_path)
    user_simulator_agent = instantiate(cfg.user_simulator_agent)

    valid_tapes: list[FormFillerTape] = []
    for tape in tapes:
        if user_simulator_agent.can_continue(tape):
            valid_tapes.append(tape)
        else:
            logger.debug(f"Skipping tape {tape.metadata.id} as it is not continuable")
    logger.info(f"Continuable tapes: {len(valid_tapes)}/{len(tapes)} total tapes")

    predicted_tapes: list[FormFillerTape] = []
    failed_tapes = []
    user_simulator_tapes = []
    for i, (exception, form_filler_tape, user_simulator_tape) in enumerate(
        lazy_thread_pool_processor(
            stream=tqdm(
                valid_tapes,
                total=min(len(valid_tapes), cfg.max_continuable_tapes)
                if cfg.max_continuable_tapes >= 0
                else len(valid_tapes),
                desc="Running user simulator agent",
            ),
            worker_func=lambda tape: run_user_simulator_agent(tape, user_simulator_agent),
            n_workers=cfg.n_workers,
        )
    ):
        if cfg.max_continuable_tapes >= 0 and i >= cfg.max_continuable_tapes:
            logger.warning(f"Reached max_continuable_tapes limit of {cfg.max_continuable_tapes}")
            break

        user_simulator_tapes.append(user_simulator_tape)

        if exception is not None:
            logger.error(f"Tape {form_filler_tape.metadata.id} failed with error: {exception}")
            logger.exception(exception, exc_info=exception)
            failed_tapes.append(form_filler_tape)

        else:
            logger.debug(
                f"Successfully continued tape {form_filler_tape.metadata.id} with {form_filler_tape.metadata.n_added_steps} steps"
            )
            predicted_tapes.append(form_filler_tape)

    counters = {
        "total_input_tapes": len(tapes),
        "total_continuable_tapes": len(valid_tapes),
        "total_user_simulator_tapes": len(user_simulator_tapes),
        "total_predicted_tapes": len(predicted_tapes),
        "total_failed_tapes": len(failed_tapes),
    }
    logger.info(json.dumps(counters, indent=2))
    with open(output_path / "counters.json", "w") as f:
        json.dump(counters, f, indent=2)

    with stream_yaml_tapes(output_path / "user_simulator_tapes.yaml") as saver:
        for tape in tqdm(user_simulator_tapes, desc="Saving user simulator tapes"):
            saver.save(tape)

    with stream_yaml_tapes(output_path / "user_predicted_tapes.yaml") as saver:
        for tape in tqdm(predicted_tapes, desc="Saving predicted tapes"):
            saver.save(tape)

    if failed_tapes:
        with stream_yaml_tapes(output_path / "user_failed_tapes.yaml") as saver:
            for tape in tqdm(failed_tapes, desc="Saving failed tapes"):
                saver.save(tape)


if __name__ == "__main__":
    main()



import json
import logging
import os
import time

import hydra
from hydra.utils import instantiate
from joblib import Parallel, delayed
from omegaconf import DictConfig

from tapeagents.io import save_json_tape, save_tape_images
from tapeagents.llms import TrainableLLM
from tapeagents.orchestrator import get_agent_and_env_from_config
from tapeagents.tools.container_executor import init_code_sandbox

from ..eval import get_exp_config_dict, load_dataset, solve_task

logger = logging.getLogger(__name__)


@hydra.main(
    version_base=None,
    config_path="../../../conf",
    config_name="gaia_mcp",
)
def main(cfg: DictConfig) -> None:
    tasks = load_dataset(cfg.split)
    init_code_sandbox(cfg.exp_path)
    dt = time.perf_counter()
    n_workers = cfg.batch or 1
    tapes_dir = os.path.join(cfg.exp_path, "tapes")
    if len(cfg.only_tasks):
        tasks = cfg.only_tasks
    else:
        tasks = [
            (level, task_num)
            for level, level_tasks in tasks.items()
            for task_num, _ in enumerate(level_tasks)
            if not task_already_solved(task_num, level, tapes_dir, cfg.retry_unsolved)
        ]
    logger.info(f"Solve {len(tasks)} tasks using {n_workers} workers")
    Parallel(n_jobs=n_workers, prefer="processes")(
        [delayed(task_worker)(cfg, level, task_num) for level, task_num in tasks]
    )
    dt = time.perf_counter() - dt
    logger.info(f"Done, elapsed time: {dt:.2f} sec")


def validate_config(cfg, llm, tapes_dir):
    if os.path.exists(tapes_dir):
        old_exp_cfg = get_exp_config_dict(cfg.exp_path)
        assert (
            old_exp_cfg["llm"]["model_name"] == llm.model_name
        ), f"Exp dir model name mismatch: old {old_exp_cfg['llm']['model_name']}, new {llm.model_name}"
        assert old_exp_cfg["split"] == cfg.split, f"Exp split: old {old_exp_cfg['split']}, new {cfg.split}"
    os.makedirs(tapes_dir, exist_ok=True)


def task_already_solved(i: int, level: int, tapes_dir: str, retry_unsolved: bool) -> bool:
    tape_name = f"l{level}_task{i:03d}"
    tape_path = os.path.join(tapes_dir, f"{tape_name}.json")
    result = None
    tape_exists = os.path.exists(tape_path)
    solved = tape_exists
    if retry_unsolved and tape_exists:
        with open(tape_path) as f:
            tape_dict = json.load(f)
        result = tape_dict["metadata"]["result"]
        solved = result not in ["", None, "None", "none", "null"]
        if not solved:
            old_file_idx = 0
            while os.path.exists(f"{tape_path}.{old_file_idx}"):
                old_file_idx += 1
            os.rename(tape_path, f"{tape_path}.{old_file_idx}")

    return solved


def task_worker(cfg: DictConfig, level: int, task_num: int):
    os.makedirs(os.path.join(cfg.exp_path, "logs"), exist_ok=True)
    log_file = os.path.join(cfg.exp_path, "logs", f"evaluate.{os.getpid()}.log")
    log_handler = logging.FileHandler(log_file)
    log_handler.setLevel(logging.INFO)
    logging.basicConfig(
        format="%(asctime)s - PID_%(process)d - Thread_%(threadName)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
        handlers=[log_handler, logging.StreamHandler()],
        force=True,  # forget previous handlers
    )
    timers = {}

    t = time.perf_counter()
    tasks = load_dataset(cfg.split)
    task = tasks[level][task_num]
    timers["load_task"] = time.perf_counter() - t

    t = time.perf_counter()
    llm: TrainableLLM = instantiate(cfg.llm)
    timers["instantiate_llm"] = time.perf_counter() - t

    os.environ["TAPEAGENTS_SQLITE_DB"] = os.path.join(cfg.exp_path, "tapedata.sqlite")
    tapes_dir = os.path.join(cfg.exp_path, "tapes")
    validate_config(cfg, llm, tapes_dir)
    images_dir = os.path.join(cfg.exp_path, "attachments", "images")
    os.makedirs(images_dir, exist_ok=True)

    t = time.perf_counter()
    agent, env = get_agent_and_env_from_config(cfg)
    timers["create_agent_env"] = time.perf_counter() - t

    tape = solve_task(task, agent, env, level, task_num, tapes_dir)

    t = time.perf_counter()
    env.close()
    timers["close_env"] = time.perf_counter() - t

    tape.metadata.other["timers"] |= timers
    save_json_tape(tape, tapes_dir, f"l{level}_task{task_num:03d}")
    save_tape_images(tape, images_dir)
    logger.info(f"Saved to {tapes_dir}")


if __name__ == "__main__":
    main()



"""
Optimizable llm functions.
"""

from __future__ import annotations

import logging
import re
from typing import Any, Generator, Type

from pydantic import BaseModel

from tapeagents.agent import Agent, Node
from tapeagents.core import Prompt, Step, Tape
from tapeagents.dialog_tape import (
    AssistantStep,
    AssistantThought,
)
from tapeagents.llms import LLMStream
from tapeagents.tool_calling import FunctionCall, ToolCall, ToolCalls

logger = logging.getLogger(__name__)

LLM_FUNCTION_TEMPLATE = """{function_desc}
{partial_demos}
---

Follow the following format.

{format_desc}
{demos}
---

{input_prompt}"""


class Variable(BaseModel):
    """Base class for all LLM function inputs and outputs"""

    name: str
    prefix: str = ""
    desc: str = ""
    separator: str = " "

    def get_prefix(self):
        return self.prefix or f"{self.name.title()}:"

    def get_desc(self):
        return self.desc or f"${{{self.name}}}"

    def render(self, value: Step):
        return value.content  # type: ignore


class Input(Variable):
    """Describes an input to an LLM-based function."""

    pass


class Output(Variable):
    """Describes an output of an LLM-based function."""

    def parse(self, text: str) -> Step:
        raise NotImplementedError()


class AssistantOutput(Output):
    """Describes an output of an LLM-based function."""

    def parse(self, text: str) -> Step:
        return AssistantStep(content=text)


class ThoughtOutput(Output):
    def parse(self, text: str):
        return AssistantThought(content=text)


class ToolCallOutput(Output):
    tool_name: str
    arg_name: str

    def parse(self, text: str):
        tc = ToolCall(function=FunctionCall(name=self.tool_name, arguments={self.arg_name: text}))
        return ToolCalls(tool_calls=[tc])

    def render(self, value: ToolCalls):
        return value.tool_calls[0].function.arguments[self.arg_name]


class ReasoningOutput(ThoughtOutput):
    @classmethod
    def for_output(cls, output_name: str):
        return cls(
            name="reasoning",
            prefix="Reasoning: Let's think step by step in order to",
            desc=f"""${{produce the {output_name}}}. We ...""",
        )


class LLMFunctionTemplate(BaseModel):
    desc: str
    inputs: list[Input]
    outputs: list[Output]
    partial_demos: list[dict] = []
    demos: list[dict] = []

    def make_prompt(self, input_values: list) -> Prompt:
        def render_demo(demo: dict) -> str:
            lines = []
            for input in self.inputs:
                if input.name in demo:
                    lines.append(f"{input.get_prefix()}{input.separator}{input.render(demo[input.name])}")
            for output in self.outputs:
                if output.name in demo:
                    lines.append(f"{output.get_prefix()}{output.separator}{output.render(demo[output.name])}")
            return "\n".join(lines)

        # PARTIAL DEMOS
        blocks = [render_demo(demo) for demo in self.partial_demos]
        partial_demos = "\n---\n\n" + "\n\n".join(blocks) + "\n" if blocks else ""

        # FORMAT DESCRIPTION
        lines = []
        for input in self.inputs:
            lines.append(f"{input.get_prefix()} {input.get_desc()}")
        for output in self.outputs:
            lines.append(f"{output.get_prefix()} {output.get_desc()}")
        format_desc = "\n".join(lines)

        # DEMOS
        blocks = [render_demo(demo) for demo in self.demos]
        demos = "\n---\n\n" + "\n\n---\n\n".join(blocks) + "\n" if blocks else ""

        # INPUT PROMPT
        lines = []
        for input, value in zip(self.inputs, input_values):
            lines.append(f"{input.get_prefix()}{input.separator}{input.render(value)}")
        lines.append(self.outputs[0].get_prefix())
        input_prompt = "\n".join(lines)

        text = LLM_FUNCTION_TEMPLATE.format(
            function_desc=self.desc,
            partial_demos=partial_demos,
            format_desc=format_desc,
            demos=demos,
            input_prompt=input_prompt,
        )
        return Prompt.from_user_message(text)

    def generate_steps(self, agent: Agent, tape: Tape, llm_stream: LLMStream) -> Generator[Step]:
        # TODO: streaming
        output_values = []
        output_text = llm_stream.get_text()
        for i, output in enumerate(self.outputs):
            if not isinstance(output, (ThoughtOutput, ToolCallOutput, AssistantOutput)):
                raise NotImplementedError(f"Output type {output} not implemented")
            # Find the variable output.name in the llm output
            values = re.split(f"{output.name}:", output_text, flags=re.IGNORECASE)
            if len(values) == 2:
                # Variable output.name found in llm output
                value = values[1].split("\n")[0].strip()  # heuristic for the end of the argument
                if isinstance(output, ReasoningOutput):
                    # if present, remove the prefix from the reasoning output
                    values = re.split(output.prefix, output_text, flags=re.IGNORECASE)
                    if len(values) == 2:
                        value = values[1].split("\n")[0].strip()  # heuristic for the end of the argument
                    else:
                        logger.debug(f"No prefix to remove from reasoning output key {output.name}")
            elif len(values) == 1:
                # Variable output.name not found in llm output
                # Set value to empty string to keep positional order
                value = ""
                if i == 0:
                    # For the first output, llm output might not repeat prefix
                    values = re.split(r"^[A-z]+:", output_text, flags=re.IGNORECASE)
                    if len(values) == 1:
                        # llm output doesn't start with another variable output
                        # we assume the first section of the llm output is the value
                        value = output_text.split("\n")[0].strip()  # heuristic for the end of the output
                        logger.debug(
                            f"Assuming value for the output key '{output.name}' is the first llm output section: '{value}'"
                        )
                    else:
                        # llm output started with another variable output
                        # Not throwing an error here because the first (reasoning) output might not be present be the second (query/answer) output is
                        logger.error(
                            f"Could not assume value for the output key '{output.name}' from the first llm output section: '{values[0]}'"
                        )
                else:
                    raise ValueError(f"Could not find output key '{output.name}' in output_text: '{output_text}'")
            else:
                raise ValueError(
                    f"Found multiple instances of output key `{output.name}` in output_text: {output_text}"
                )
            output_values.append(value)

        if len(output_values) != len(self.outputs):
            logger.warning(f"Could not find all outputs in output_text: {output_text}\nOutput found: {output_values}")

        for output, value in zip(self.outputs, output_values):
            yield output.parse(value)


class KindRef(BaseModel):
    """Refer to the input by the step kind. Refers the last step with the given kind."""

    kind: str


def by_step(step_class: Type) -> KindRef:
    return KindRef(kind=step_class.get_kind())


class NodeRef(BaseModel):
    name: str


def by_node(node: Node) -> NodeRef:
    return NodeRef(name=node.name)


class LLMFunctionNode(Node):
    template_name: str
    input_refs: list[int | NodeRef | KindRef | Step] = []

    def get_function(self, agent):
        template = agent.templates[self.template_name]
        if not isinstance(template, LLMFunctionTemplate):
            raise ValueError(f"Template {self.template_name} is not an LLMFunctionTemplate")
        return template

    def extract_inputs(self, tape: Tape, index: int | None = None):
        if index is None:
            index = len(tape) + 1

        inputs = []
        for ref in self.input_refs:
            resolved = False
            steps = tape.steps[:index]
            if isinstance(ref, int):
                inputs.append(steps[ref])
            elif isinstance(ref, NodeRef):
                for step in reversed(steps):
                    if step.metadata.node == ref.name:
                        inputs.append(step)
                        resolved = True
                        break
                if not resolved:
                    raise ValueError(f"Node {ref.name} not found in tape")
            elif isinstance(ref, KindRef):
                for step in reversed(steps):
                    if step.kind == ref.kind:
                        inputs.append(step)
                        resolved = True
                        break
                if not resolved:
                    raise ValueError(f"Step with kind {ref.kind} not found in tape")
            elif isinstance(ref, Step):
                inputs.append(ref)
            else:
                raise ValueError(f"Invalid input reference {ref}")
        return inputs

    def make_prompt(self, agent, tape: Tape):
        return self.get_function(agent).make_prompt(self.extract_inputs(tape))

    def generate_steps(self, agent: Any, tape: Tape, llm_stream: LLMStream):
        yield from self.get_function(agent).generate_steps(agent, tape, llm_stream)

    def extract_demo(self, agent: Agent, tape: Tape, index: int):
        func = self.get_function(agent)
        input_values = self.extract_inputs(tape, index)
        output_values = tape.steps[index : index + len(func.outputs)]
        demo = {}
        for input, value in zip(func.inputs, input_values):
            demo[input.name] = value
        for output, value in zip(func.outputs, output_values):
            demo[output.name] = value
        return demo



"""
Various utility functions.
"""

import base64
import difflib
import fcntl
import importlib
import io
import json
import os
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Any

import jsonref
from PIL import Image
from pydantic import TypeAdapter
from termcolor import colored


# Custom exception for fatal errors
class FatalError(Exception):
    pass


def json_value_from_str(v: Any) -> Any:
    if isinstance(v, str):
        try:
            v = json.loads(v)
        except Exception:
            pass
    return v


def diff_dicts(a: dict, b: dict) -> str:
    str_a = json.dumps(a, indent=2, ensure_ascii=False, sort_keys=True)
    str_b = json.dumps(b, indent=2, ensure_ascii=False, sort_keys=True)
    return diff_strings(str_a, str_b, by_words=True)


def diff_strings(a: str, b: str, use_html: bool = False, by_words: bool = False) -> str:
    def html_colored(s, on_color):
        color = on_color[3:] if on_color.startswith("on_") else on_color
        style = f"color: {color} !important;"
        if color == "red":
            style += "text-decoration: line-through !important;"  # strike-through for deleted text
        return f'<span style="{style}">{s}</span>'

    color_fn = html_colored if use_html else colored
    output = []
    if by_words:
        a = a.replace("\n", " \n").split(" ")  # type: ignore
        b = b.replace("\n", " \n").split(" ")  # type: ignore
    matcher = difflib.SequenceMatcher(None, a, b)
    splitter = " " if by_words else ""
    for opcode, a0, a1, b0, b1 in matcher.get_opcodes():
        a_segment = splitter.join(a[a0:a1])
        b_segment = splitter.join(b[b0:b1])
        if opcode == "equal":
            output.append(a_segment)
        elif opcode == "insert":
            output.append(color_fn(f"{b_segment}", on_color="on_green"))
        elif opcode == "delete":
            output.append(color_fn(f"{a_segment}", on_color="on_red"))
        elif opcode == "replace":
            output.append(color_fn(f"{b_segment}", on_color="on_green"))
            output.append(color_fn(f"{a_segment}", on_color="on_red"))
    return splitter.join(output)


def sanitize_json_completion(completion: str) -> str:
    """
    Return only content inside the first pair of triple backticks if they are present.
    """
    tiks_counter = 0
    lines = completion.strip().split("\n")
    clean_lines = []
    for line in lines:
        if line.startswith("```"):
            tiks_counter += 1
            if tiks_counter == 1:
                clean_lines = []
            elif tiks_counter == 2:
                break
            continue
        clean_lines.append(line)
    return "\n".join(clean_lines)


def without(d: dict, key: str) -> dict:
    if key in d:
        d.pop(key)
    return d


def get_step_schemas_from_union_type(cls, simplify: bool = True) -> str:
    schema = TypeAdapter(cls).json_schema()
    dereferenced_schema: dict = dict(jsonref.replace_refs(schema, proxies=False))  # type: ignore
    clean_schema = []
    for step in dereferenced_schema["oneOf"]:
        step["properties"] = without(step["properties"], "metadata")
        if simplify:
            step = without(step, "title")
            for prop in step["properties"]:
                step["properties"][prop] = without(step["properties"][prop], "title")
            step["properties"]["kind"] = {"const": step["properties"]["kind"]["const"]}
        clean_schema.append(step)
    return json.dumps(clean_schema, ensure_ascii=False)


def image_base64_message(image_path: str) -> dict:
    image_extension = os.path.splitext(image_path)[1][1:]
    content_type = f"image/{image_extension}"
    base64_image = encode_image(image_path)
    message = {"type": "image_url", "image_url": {"url": f"data:{content_type};base64,{base64_image}"}}
    return message


def resize_base64_message(message: dict, max_size: int = 1280) -> dict:
    base64_image = message["image_url"]["url"].split(",", maxsplit=1)[1]
    image = Image.open(io.BytesIO(base64.b64decode(base64_image)))
    if image.size[0] > max_size or image.size[1] > max_size:
        image.thumbnail((max_size, max_size), Image.Resampling.LANCZOS)
    with tempfile.NamedTemporaryFile() as tmp:
        image.save(tmp.name, "PNG")
        base64_image = encode_image(tmp.name)
        content_type = "image/png"
    message = {"type": "image_url", "image_url": {"url": f"data:{content_type};base64,{base64_image}"}}
    return message


def encode_image(image_path):
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")


@contextmanager
def acquire_timeout(lock, timeout):
    result = lock.acquire(timeout=timeout)
    try:
        yield result
    finally:
        if result:
            lock.release()


def class_for_name(full_name: str) -> Any:
    if "." in full_name:
        module_name, class_name = full_name.rsplit(".", 1)
    else:
        module_name = "."
        class_name = full_name
    # load the module, will raise ImportError if module cannot be loaded
    m = importlib.import_module(module_name)
    # get the class, will raise AttributeError if class cannot be found
    c = getattr(m, class_name)
    return c


class Lock:
    def __init__(self, name: str):
        self.name = f"./.{name}.lock"
        Path(self.name).touch()

    def __enter__(self):
        self.fp = open(self.name)
        fcntl.flock(self.fp.fileno(), fcntl.LOCK_EX)

    def __exit__(self, _type, value, tb):
        fcntl.flock(self.fp.fileno(), fcntl.LOCK_UN)
        self.fp.close()



import json
import logging
import os
import shutil
import subprocess
import time
from typing import Any, Counter

import yaml
from huggingface_hub import snapshot_download
from pdf2image import convert_from_path
from termcolor import colored

from tapeagents.agent import Agent
from tapeagents.environment import ToolCollectionEnvironment
from tapeagents.io import load_tapes, save_json_tape
from tapeagents.orchestrator import main_loop
from tapeagents.renderers import step_view
from tapeagents.tools.code_executor import PythonCodeAction
from tapeagents.tools.simple_browser import SimpleTextBrowser
from tapeagents.tools.web_search import SearchAction

from .scorer import question_scorer
from .steps import GaiaAnswer, GaiaMetadata, GaiaQuestion, GaiaTape, ImageObservation

logger = logging.getLogger(__name__)

DATASET_DIR = "data/gaia/"


def get_git_revision_hash() -> str:
    return subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("ascii").strip()


def tape_correct(tape: GaiaTape) -> bool:
    if not tape.metadata.result:
        return False
    predicted = str(tape.metadata.result)
    golden = tape.metadata.task["Final answer"]
    if golden == "?":  # placeholder for hidden answer in test set
        return False
    return question_scorer(predicted, golden)


def tape_without_result(tape: GaiaTape) -> bool:
    return tape.metadata.result in ["", "None", None]


def calculate_accuracy(tapes: list[GaiaTape], show_intermediate=False, show_wrong=False):
    accs = []
    accuracy = 0.0
    for tape in tapes:
        correct = tape_correct(tape)
        if show_wrong and not correct:
            for step in tape:
                print("-" * 80)
                print(step_view(step))
            print("=" * 120)

        accs.append(int(correct))
        accuracy = sum(accs) * 100 / len(accs)
        if show_intermediate:
            print(
                tape.metadata.task["Final answer"],
                "|",
                colored(str(tape.metadata.result), "green" if correct else "red"),
            )
            print(f"{len(accs)}: Accuracy {accuracy:.2f}")
    return accuracy, sum(accs)


def majority_vote(results: list[Any]) -> int:
    result_counts = Counter(results)
    top = result_counts.most_common()
    best, _ = top.pop(0)
    while len(top) and best in ["", "None", None]:
        best, _ = top.pop(0)
    best_idx = [i for i, result in enumerate(results) if result == best][0]
    return best_idx


def download_dataset():
    logger.info("Downloading GAIA dataset...")
    repo = "gaia-benchmark/GAIA"
    os.makedirs(DATASET_DIR, exist_ok=True)
    snapshot_download(repo_id=repo, repo_type="dataset", local_dir=DATASET_DIR, local_dir_use_symlinks=False)


def load_dataset(split: str):
    tasks = {1: [], 2: [], 3: []}
    fname = os.path.join(DATASET_DIR, "2023", split, "metadata.jsonl")
    if not os.path.exists(fname):
        download_dataset()
    if not os.path.exists(fname):
        raise FileNotFoundError(f"Dataset not found: {fname}")
    with open(fname) as f:
        for line in f:
            task = json.loads(line)
            if task["file_name"]:
                task["file_name"] = os.path.join(DATASET_DIR, "2023", split, task["file_name"])
            tasks[task["Level"]].append(task)

    logger.info(f"GAIA Tasks: Level 1: {len(tasks[1])}, Level 2: {len(tasks[2])}, Level 3: {len(tasks[3])}")
    return tasks


def solve_task(
    task: dict,
    agent: Agent,
    env: ToolCollectionEnvironment,
    level: int,
    task_num: int,
    tapes_dir: str,
    max_loops: int = 50,
    max_action_repetitions: int = 3,
) -> GaiaTape:
    start_steps = task_to_observations(task)
    t = time.perf_counter()
    tape = GaiaTape(steps=start_steps)
    loop_timout_sec = 30 * 60
    tmp_file = os.path.join(tapes_dir, f"l{level}_task{task_num:03d}.json.tmp")
    try:
        start_time = time.perf_counter()
        for event in main_loop(agent, tape, env, max_loops=max_loops):
            if time.perf_counter() - start_time > loop_timout_sec:
                tape.metadata.error = "Timeout, task took too long"
                logger.warning("Timeout, task took too long")
                break
            if partial_tape := (event.agent_tape or event.env_tape):
                tape = partial_tape
                tape.metadata = GaiaMetadata.model_validate(tape.metadata.model_dump() | {"task": task, "level": level})
                save_json_tape(tape, tmp_file)
            if action_repetitions(tape) >= max_action_repetitions:
                break
    except Exception as e:
        tape.metadata.error = str(e)
        logger.exception(f"Failed to solve task: {e}")
    result = tape[-1].answer if isinstance(tape[-1], GaiaAnswer) else None  # type: ignore
    result = str(result) if result is not None else ""
    logger.info(f"Expected: {task['Final answer']}, Agent produced: {result}")
    tape.metadata = GaiaMetadata.model_validate(
        tape.metadata.model_dump() | {"task": task, "result": result, "level": level}
    )
    tape.metadata.other["timers"] = {"solve_task": time.perf_counter() - t}
    if os.path.exists(tmp_file):
        os.unlink(tmp_file)
    return tape


def action_repetitions(tape: GaiaTape) -> int:
    unique_actions = {}
    for step in tape:
        if isinstance(step, (SearchAction, PythonCodeAction)):
            key = step.llm_view()
            unique_actions[key] = unique_actions.get(key, 0) + 1
    return max(unique_actions.values(), default=0)


def ensemble_results(all_tapes: list[list[GaiaTape]], oracle: bool = False) -> list[GaiaTape]:
    ensemble = []
    improved = 0
    degraded = 0
    added_result = 0
    no_result = 0
    for i, tapes in enumerate(zip(*all_tapes)):
        tapes: list[GaiaTape] = tapes
        results = [tape.metadata.result for tape in tapes]
        most_common_idx = majority_vote(results)
        if oracle:
            for j, tape in enumerate(tapes):
                if tape_correct(tape):
                    most_common_idx = j
                    break
        best_tape = tapes[most_common_idx].copy()

        tape0 = tapes[0]
        tape0_correct = int(tape_correct(tape0))
        ensemble_correct = int(tape_correct(best_tape))
        expected = tapes[0].metadata.task["Final answer"]
        change = "switched" if best_tape.metadata.result != results[0] else "same"
        log_message = f"{i+1}: {tape0_correct} -> {ensemble_correct} | {change} {most_common_idx+1} ({best_tape.metadata.result}) of {results}. Expected: {expected}"
        if tape0_correct < ensemble_correct:
            logger.info("Improved")
            improved += 1
        elif tape0_correct > ensemble_correct:
            logger.info("Degraded")
            degraded += 1
        if tape_without_result(best_tape):
            no_result += 1
        if tape_without_result(tape0) and not tape_without_result(best_tape):
            added_result += 1
            logger.info("Added result")
        logger.info(log_message)
        ensemble.append(best_tape)
    logger.info(f"Improved {improved}, degraded {degraded}, no result {no_result}, added result {added_result}")
    return ensemble


def ensemble_files(tape_dirs: list[str], out_dir: str = ""):
    tapes: list[list[GaiaTape]] = [load_tapes(GaiaTape, tape_dir, file_extension=".json") for tape_dir in tape_dirs]  # type: ignore
    ensembled_tapes = ensemble_results(tapes)
    acc, num = calculate_accuracy(ensembled_tapes, show_intermediate=False)
    logger.info(f"Ensembled {len(tape_dirs)} accuracy: {acc:.2f} ({num} of {len(ensembled_tapes)})")
    if out_dir:
        out_tapes_dir = os.path.join(out_dir, "tapes")
        os.makedirs(out_tapes_dir)
        for i, tape in enumerate(ensembled_tapes):
            save_json_tape(tape, out_tapes_dir, f"tape{i+1:03d}")
        with open(os.path.join(out_dir, "ensemble_config.json"), "w") as f:
            json.dump({"sources": tape_dirs, "mode": "majority"}, f, indent=2)
        logger.info(f"Saved to {out_dir}")


def get_exp_config_dict(exp_path):
    config_path = os.path.join(exp_path, ".hydra", "config.yaml")
    assert os.path.exists(config_path), f"Config file {config_path} not found"
    with open(config_path) as f:
        cfg = yaml.safe_load(f)
    return cfg


def pdf_to_images(filename: str, n_pages: int = 3):
    images = []
    for i, image in enumerate(convert_from_path(filename)):
        page_index = i + 1
        page_fname = filename[:-4] + f"_{page_index}.png"
        if os.path.exists(page_fname):
            images.append(page_fname)
            continue
        image.save(page_fname)
        images.append(page_fname)
    return images[:n_pages], len(images)


def task_to_observations(task: dict, max_doc_length: int = 8000) -> list[GaiaQuestion | ImageObservation]:
    logger.info(f"Question: {task['Question']}")
    browser = SimpleTextBrowser()
    steps: list[GaiaQuestion | ImageObservation] = [GaiaQuestion.from_task(task)]
    filename: str | None = steps[0].filename
    steps[0].filename = None
    if filename:
        name, ext = filename.rsplit(".", maxsplit=1)
        ext = ext.lower()
        if ext == "zip":
            folder_name = name
            os.makedirs(folder_name, exist_ok=True)
            shutil.unpack_archive(filename, folder_name)
            document_text = "\n\nArchive contains the following files:\n"
            for i, file in enumerate(os.listdir(folder_name)):
                file_path = os.path.join(folder_name, file)
                content = browser.get_whole_document(file_path)
                file_text = f"{i+1}. {file}. Content:\n{content}\n\n"
                if len(file_text) > max_doc_length:
                    file_text = ""
                file_text += f"{i+1}. Path to the '{file}': {file_path}"
                document_text += file_text
        elif ext in ("png", "jpg", "jpeg"):
            steps.append(ImageObservation(image_path=filename, image_caption="Attached image"))
            document_text = ""
        else:
            attach_doc_text = True
            if ext == "pdf":
                images, total_pages = pdf_to_images(filename)
                if total_pages <= 3:
                    attach_doc_text = False
                for i, img_path in enumerate(images):
                    steps.append(ImageObservation(image_path=img_path, image_caption=f"PDF page {i+1}"))
            if attach_doc_text:
                try:
                    content = browser.get_whole_document(filename)
                except Exception as e:
                    logger.exception(f"Failed to read document: {e}")
                    content = ""
                document_text = f"\n\nAttached {ext.upper()} file content:\n{content}\n"
                if not len(content) or len(document_text) > max_doc_length:
                    document_text = ""
            else:
                document_text = "\nDocument pages attached as images below"
            steps[0].filename = filename
        steps[0].content += document_text  # type: ignore
    return steps



import logging
import os
import sys
from pathlib import Path

import transformers
from omegaconf import DictConfig

from tapeagents.finetune.data import load_samples
from tapeagents.io import load_tapes
from tapeagents.llms import LLM, LLMCall, ReplayLLM, TrainableLLM
from tapeagents.orchestrator import replay_tapes

sys.path.append(str(Path(__file__).parent.parent.resolve()))  # allow to import from examples

from examples.gsm8k_tuning.finetune_student import get_training_samples_from_tapes
from examples.gsm8k_tuning.math_agent import MathAgent, MathTape
from examples.rl_gsm8k.orchestrate_rl import CoTMathAgent, RLMathTape, extract_tape_training_samples

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

res_path = Path(__file__).parent.resolve() / "res"


def mock_llm(run_dir: str) -> LLM:
    llama = TrainableLLM(
        base_url="https://api.together.xyz",
        model_name="meta-llama/Meta-Llama-3-70B-Instruct-Turbo",
        tokenizer_name="meta-llama/Meta-Llama-3-70B-Instruct",
        parameters=dict(temperature=0.7, max_tokens=512),
    )
    return ReplayLLM.from_llm(llama, run_dir)


def test_gsm8k_tuning_tapes_generation():
    run_dir = f"{res_path}/gsm8k_tuning"
    llm = mock_llm(run_dir)
    agent = MathAgent.create(llm)
    tapes = load_tapes(MathTape, os.path.join(run_dir, "tapes"), file_extension=".json")
    logger.info(f"Validate {len(tapes)} tapes")
    fails = replay_tapes(agent, tapes, reuse_observations=True)
    assert fails == 0, f"{fails} failed tapes"


def test_gsm8k_tuning_samples_prep():
    run_dir = f"{res_path}/gsm8k_tuning"
    training_samples = load_samples(f"{run_dir}/training_samples.jsonl")
    new_training_samples = get_training_samples_from_tapes(f"{run_dir}/tapes/")
    assert training_samples == new_training_samples


def test_rl_gsm8k_data():
    run_dir = f"{res_path}/rl_gsm8k"
    tapes = load_tapes(RLMathTape, run_dir, file_extension=".json")
    llm = mock_llm(run_dir)
    llm.tokenizer = transformers.AutoTokenizer.from_pretrained("meta-llama/Llama-3.1-8B-Instruct")
    llm.parameters = {"max_tokens": 10000}
    agent = CoTMathAgent.create(system_prompt="", llm=llm, max_prompt_length=1024)
    cfg = DictConfig(
        {"dataset_name": "math", "llm": {"parameters": {"max_tokens": 2048}}, "finetune": {"seq_length": 2048}}
    )
    training_samples = []
    for tape in tapes:
        for step in tape:
            if llm_call_data := step.metadata.other.get("llm_call"):
                step.metadata.other["llm_call"] = LLMCall(**llm_call_data)
        training_sample, _ = extract_tape_training_samples(tape, agent, cfg)
        training_samples.append(training_sample[0])
    new_training_samples = load_samples(f"{run_dir}/training_samples.jsonl")
    assert training_samples == new_training_samples


if __name__ == "__main__":
    test_gsm8k_tuning_tapes_generation()
    test_gsm8k_tuning_samples_prep()
    test_rl_gsm8k_data()



import hashlib
import json
import logging
import os
import threading
from abc import abstractmethod
from typing import Generator

from termcolor import colored

from tapeagents.config import common_cache_dir
from tapeagents.core import LLMOutput, Prompt
from tapeagents.llms.base import LLM, LLMEvent, LLMStream, closest_prompt
from tapeagents.observe import retrieve_all_llm_calls
from tapeagents.utils import diff_strings

logger = logging.getLogger(__name__)
# Use this variable to force all LLMs to use cache from the sqlite DB
# This is meant to be used for testing purposes only
_REPLAY_SQLITE: str = ""


class CachedLLM(LLM):
    """A caching wrapper for LLM implementations that stores and retrieves previous LLM responses.

    This class implements caching functionality for LLM responses to avoid redundant API calls
    and enable replay of previous interactions. It supports both file-based caching and SQLite-based
    replay functionality for testing purposes.

    Attributes:
        use_cache (bool): Flag to enable/disable caching functionality. Defaults to False.
        stream (bool): Flag to enable/disable streaming responses. Defaults to False.
        _cache (dict): Internal cache storage mapping prompt keys to LLM responses.

    The cache can be initialized in two modes:
    1. SQLite replay mode: Used for testing, enforces cache hits only
    2. File-based cache mode: Stores responses in a jsonl file for persistence

    Cache keys are generated based on the prompt content, excluding the prompt ID.
    During testing (replay mode), exact text matching is used instead of hashing.
    """

    use_cache: bool = False
    stream: bool = False
    _cache: dict = {}

    def model_post_init(self, __content):
        if _REPLAY_SQLITE:
            self.use_cache = True
            self._cache = {}
            llm_calls = retrieve_all_llm_calls(_REPLAY_SQLITE)
            for llm_call in llm_calls:
                key = self.get_prompt_key(llm_call.prompt)
                self._cache[key] = [LLMEvent(output=llm_call.output)]
            logger.info(f"Enforced LLM cache from {_REPLAY_SQLITE}, {len(self._cache)} entries")
            return
        elif not self.use_cache:
            return
        param_hash = self._key(json.dumps({k: v for k, v in self.parameters.items() if k != "token"}))
        name = self.model_name.replace("/", "__")
        prefix = f"llm_cache_{name}_{param_hash}."
        cache_dir = common_cache_dir()
        self._cache_file = os.path.join(cache_dir, f"{prefix}{os.getpid()}.{threading.get_native_id()}.jsonl")
        if os.path.exists(cache_dir):
            for fname in os.listdir(cache_dir):
                if not fname.startswith(prefix):
                    continue
                with open(os.path.join(cache_dir, fname)) as f:
                    for line in f:
                        key, event_dict = json.loads(line)
                        if key not in self._cache:
                            self._cache[key] = []
                        self._cache[key].append(event_dict)
            logger.info(f"Loaded {len(self._cache)} llm calls from cache {cache_dir}")
        else:
            logger.info(f"Cache dir {cache_dir} does not exist")

    def reindex_log(self):
        """
        Reindex the log data into cache.

        This method iterates through the log entries, validates each prompt and output,
        and adds them to the cache using the prompt key as index. Each entry is converted
        to an LLMEvent model before caching.

        Side Effects:
            - Updates the internal cache with log data
            - Logs the total number of reindexed entries at INFO level
        """
        cnt = 0
        for log_data in self._log:
            key = self.get_prompt_key(Prompt.model_validate(log_data["prompt"]))
            self._add_to_cache(key, LLMEvent(output=LLMOutput.model_validate(log_data["output"])).model_dump())
            cnt += 1
        logger.info(f"Reindexed {cnt} log entries")

    def _add_to_cache(self, key: str, event_dict: dict):
        if not self.use_cache:
            return
        if key not in self._cache:
            self._cache[key] = []
        self._cache[key].append(event_dict)
        with open(self._cache_file, "a") as f:
            f.write(json.dumps((key, event_dict), ensure_ascii=False) + "\n")

    def get_prompt_key(self, prompt: Prompt) -> str:
        prompt_text = json.dumps(prompt.model_dump(exclude={"id"}), ensure_ascii=False, sort_keys=True)
        return self._key(prompt_text)

    def _key(self, text: str) -> str:
        if _REPLAY_SQLITE:
            # use exact text as a key during testing
            return text
        return hashlib.md5(text.encode("utf-8")).hexdigest()

    def generate(self, prompt: Prompt, **kwargs) -> LLMStream:
        """Generate a response stream from the language model based on the given prompt.

        This method handles both cached and new responses, implementing a caching mechanism
        for LLM responses to avoid redundant API calls.

        Args:
            prompt (Prompt): The prompt object containing messages to send to the LLM.
            **kwargs (dict, optional): Additional arguments to pass to the underlying LLM implementation.

        Returns:
            LLMStream: A stream of LLM events containing the model's response.

        Raises:
            ValueError: If cache miss occurs when replay mode is enabled (_REPLAY_SQLITE is True).

        Notes:
            - If caching is enabled and the prompt exists in cache, returns cached response
            - If generating new response, tokens are counted and added to total token count
            - All generated events are cached for future use if caching is enabled
            - Output is logged through the logging system
        """

        def _implementation():
            key = self.get_prompt_key(prompt)
            if self.use_cache and key in self._cache:
                logger.debug(colored(f"LLM cache hit, {len(self._cache[key])} events", "green"))
                for event_dict in self._cache[key]:
                    event = LLMEvent.model_validate(event_dict)
                    if event.output is not None:
                        self.log_output(prompt, event.output, cached=True)
                    yield event
            else:
                if _REPLAY_SQLITE:
                    closest, score = closest_prompt(key, list(self._cache.keys()))
                    logger.error(
                        f"LLM cache miss, closest in cache has score {score:.3f}\nNEW:\n{key}\nCLOSEST OLD:\n{closest}\nDIFF:\n{diff_strings(key, closest)}"
                    )
                    raise ValueError(f"LLM cache miss not allowed. Prompt key: {key}")
                toks = self.count_tokens(prompt.messages)
                self.token_count += toks
                logger.debug(f"{toks} prompt tokens, total: {self.token_count}")
                for event in self._generate(prompt, **kwargs):
                    self._add_to_cache(key, event.model_dump())
                    # note: the underlying LLM will log the output
                    yield event

        return LLMStream(_implementation(), prompt)

    def quick_response(self, text_prompt: str) -> str:
        prompt = Prompt(messages=[{"role": "user", "content": text_prompt}])
        outputs = []
        for e in self.generate(prompt):
            if e.output:
                outputs.append(e.output.content)
        return "".join(outputs)

    @abstractmethod
    def _generate(self, prompt: Prompt, **kwargs) -> Generator[LLMEvent, None, None]:
        pass



import contextlib
import gzip
import json
import logging
import os
import shutil
import sys
import tempfile
from pathlib import Path

import yaml
from omegaconf import DictConfig

from tapeagents.config import ATTACHMENT_DEFAULT_DIR, DB_DEFAULT_FILENAME
from tapeagents.core import AgentStep, TrainingText
from tapeagents.dialog_tape import DialogTape
from tapeagents.environment import EmptyEnvironment
from tapeagents.io import load_tapes
from tapeagents.llms import LLM, LiteLLM, ReplayLLM, TrainableLLM
from tapeagents.observe import init_sqlite_if_not_exists, retrieve_tape_llm_calls
from tapeagents.orchestrator import get_agent_and_env_from_config, replay_tape, replay_tapes
from tapeagents.team import TeamTape
from tests.make_test_data import run_test_in_tmp_dir

sys.path.append(str(Path(__file__).parent.parent.parent.resolve()))  # allow to import from examples

from examples.data_science import data_science
from examples.delegate import ExampleTape, FindIrregularVerbs
from examples.delegate_stack import (
    ExampleTape as ExampleTapeStack,
    Linguist,
    make_analyze_text_chain,
)
from examples.form_filler.environment import FormFillerEnvironment
from examples.form_filler.scripts.prepare_test_assets import (
    get_teacher_agent,
    get_user_simulator_agent,
    load_teacher_input_tapes,
    load_teacher_reference_tapes,
    load_user_input_tapes,
    load_user_reference_tapes,
)
from examples.gaia_agent.steps import GaiaTape
from examples.llama_agent import LLAMAChatBot
from examples.optimize.optimize import make_agentic_rag_agent, make_env
from examples.tape_improver import tape_improver
from examples.workarena.agent import WorkArenaAgent
from examples.workarena.steps import WorkArenaTape

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

res_path = Path(__file__).parent.resolve() / "res"


def mock_llm(run_dir: str) -> LLM:
    llama = TrainableLLM(
        base_url="https://api.together.xyz",
        model_name="meta-llama/Meta-Llama-3-70B-Instruct-Turbo",
        tokenizer_name="meta-llama/Meta-Llama-3-70B-Instruct",
        parameters=dict(temperature=0.7, max_tokens=512),
    )
    return ReplayLLM.from_llm(llama, run_dir)


@contextlib.contextmanager
def set_sqlite_db_dir(dir: str):
    old_path = os.environ.get("TAPEAGENTS_SQLITE_DB")
    new_path = f"{dir}/{DB_DEFAULT_FILENAME}"
    os.environ["TAPEAGENTS_SQLITE_DB"] = new_path
    init_sqlite_if_not_exists(only_once=False)
    try:
        yield
    finally:
        if old_path is not None:
            os.environ["TAPEAGENTS_SQLITE_DB"] = old_path
        else:
            del os.environ["TAPEAGENTS_SQLITE_DB"]


def load_tape_dict(run_dir: str, fname: str = "tape.json") -> dict:
    tape_fpath = os.path.join(run_dir, fname)
    with open(tape_fpath, "r") as f:
        tape_dict = json.load(f)
    return tape_dict


def load_traces(run_dir: str, fname: str = "traces.json") -> list[TrainingText]:
    traces_fpath = os.path.join(run_dir, fname)
    with open(traces_fpath, "r") as f:
        traces = [TrainingText.model_validate(t) for t in json.load(f)]
    return traces


def test_llama_agent():
    run_dir = str(res_path / "llama_agent")
    llm = mock_llm(run_dir)
    agent = LLAMAChatBot.create(llm)
    tape = DialogTape.model_validate(load_tape_dict(run_dir))

    assert replay_tape(agent, tape, reuse_observations=True)


def test_llama_agent_traces():
    run_dir = f"{res_path}/llama_agent"
    llm = mock_llm(run_dir)
    agent = LLAMAChatBot.create(llm)
    tape = DialogTape.model_validate(load_tape_dict(run_dir))
    orig_traces = load_traces(run_dir)

    with tempfile.TemporaryDirectory() as tmp_dir:
        with set_sqlite_db_dir(tmp_dir):
            traces = agent.make_training_data(tape)
            assert len(traces) == len(orig_traces), f"Expected {len(orig_traces)} traces, got {len(traces)}"
            for trace, orig_trace in zip(traces, orig_traces):
                assert trace.prompt_text == orig_trace.prompt_text
                assert trace.output_text == orig_trace.output_text


def test_llama_agent_tape_reuse():
    data_dir = f"{res_path}/llama_agent"
    llm = mock_llm(data_dir)
    agent = LLAMAChatBot.create(llm)
    tape = DialogTape.model_validate(load_tape_dict(data_dir))
    orig_traces = load_traces(data_dir)

    with tempfile.TemporaryDirectory() as run_dir:
        with set_sqlite_db_dir(run_dir):
            reused_tape, llm_calls = agent.reuse(tape)
            for reused_step, step in zip(reused_tape, tape):
                if isinstance(step, AgentStep):
                    assert isinstance(reused_step, AgentStep)
                    assert reused_step.metadata.prompt_id != step.metadata.prompt_id
    retrieve_tape_llm_calls(reused_tape)
    traces_from_logs = [agent.make_training_text(llm_call) for llm_call in llm_calls]
    direct_traces = agent.make_training_data(tape)
    assert len(traces_from_logs) == len(
        orig_traces
    ), f"Expected {len(orig_traces)} traces from logs, got {len(traces_from_logs)}"
    for trace, orig_trace in zip(traces_from_logs, orig_traces):
        assert trace.prompt_text == orig_trace.prompt_text
        assert trace.output_text == orig_trace.output_text

    assert len(direct_traces) == len(
        orig_traces
    ), f"Expected {len(orig_traces)} direct traces, got {len(direct_traces)}"
    for trace, orig_trace in zip(direct_traces, orig_traces):
        assert trace.prompt_text == orig_trace.prompt_text
        assert trace.output_text == orig_trace.output_text


def test_gaia_agent():
    run_dir = res_path / "gaia_agent"
    db_file = run_dir / "tapedata.sqlite"
    logger.info(f"Copy {run_dir / 'tapedata.sqlite.gz'} to {db_file}")
    with gzip.open(run_dir / "tapedata.sqlite.gz", "rb") as f_in:
        with open(db_file, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    try:
        os.environ["TAPEAGENTS_CACHE_DIR"] = str(run_dir / "cache")
        with open(run_dir / "config.yaml") as f:
            cfg = DictConfig(yaml.safe_load(f))
        agent, env = get_agent_and_env_from_config(cfg)
        agent.llms["default"] = ReplayLLM.from_llm(agent.llm, run_dir)
        tapes = load_tapes(
            GaiaTape,
            run_dir / "tapes",
            file_extension=".json",
            attachment_dir=str(run_dir / ATTACHMENT_DEFAULT_DIR),
        )
        logger.info(f"Validate {len(tapes)} tapes")
        fails = replay_tapes(agent, tapes, env, reuse_observations=True, stop_on_error=True)
        assert fails == 0, f"{fails} failed tapes"
    finally:
        if os.path.exists(db_file):
            os.remove(db_file)


def test_workarena_agent():
    return
    run_dir = str(res_path / "workarena" / "guided")
    llm = ReplayLLM.from_llm(LiteLLM(model_name="mock"), run_dir)
    agent = WorkArenaAgent.create(llm)
    tapes = load_tapes(WorkArenaTape, os.path.join(run_dir, "tapes"), file_extension=".json")
    logger.info(f"Validate {len(tapes)} tapes")
    fails = replay_tapes(agent, tapes, reuse_observations=True, stop_on_error=True)
    assert fails == 0, f"{fails} failed tapes"


def test_delegate():
    run_dir = str(res_path / "delegate")
    llm = mock_llm(run_dir)
    agent = FindIrregularVerbs.create(llm)

    start_tape = ExampleTape.model_validate(load_tape_dict(run_dir, fname="start_tape.json"))
    tape = ExampleTape.model_validate(load_tape_dict(run_dir))

    replay_success = replay_tape(agent, tape, start_tape=start_tape, reuse_observations=True)
    assert replay_success, "Failed to replay tape"


def test_delegate_stack():
    run_dir = str(res_path / "delegate_stack")
    llm = mock_llm(run_dir)
    agent1 = Linguist.create(llm)
    agent2 = make_analyze_text_chain(llm)

    start_tape = ExampleTapeStack.model_validate(load_tape_dict(run_dir, fname="start_tape.json"))
    tape1 = ExampleTapeStack.model_validate(load_tape_dict(run_dir, fname="tape1.json"))
    tape2 = ExampleTapeStack.model_validate(load_tape_dict(run_dir, fname="tape2.json"))

    replay_success = replay_tape(agent1, tape1, start_tape=start_tape, reuse_observations=True)
    assert replay_success, "Failed to replay tape"
    replay_success = replay_tape(agent2, tape2, start_tape=start_tape, reuse_observations=True)
    assert replay_success, "Failed to replay tape"


def test_data_science():
    run_dir = f"{res_path}/data_science"
    llm = mock_llm(run_dir)
    agent, start_tape, env = data_science.make_world(llm, EmptyEnvironment())
    final_tape = TeamTape.model_validate(load_tape_dict(run_dir, "final_tape.json"))
    replay_success = replay_tape(agent, final_tape, start_tape=start_tape, env=env, reuse_observations=True)
    assert replay_success, "Failed to replay tape"


def test_form_filler():
    os.environ["TAPEAGENTS_MOCK_DATE"] = "2024-12-09"
    assets_dir = str(Path(__file__).parent / "res" / "form_filler")
    forms_path = str(
        Path(__file__).parent.parent.parent / "examples" / "form_filler" / "assets" / "forms" / "train" / "FlyCorp"
    )
    env = FormFillerEnvironment.from_spec(forms_path)

    teacher_agent = get_teacher_agent()
    user_agent = get_user_simulator_agent()
    teacher_mock_llm = ReplayLLM.from_llm(teacher_agent.llms["default"], assets_dir)
    teacher_agent.llms = {"default": teacher_mock_llm}

    user_mock_llm = ReplayLLM.from_llm(user_agent.llms["default"], assets_dir)
    user_agent.llms = {"default": user_mock_llm}

    # teacher_output_tapes, user_output_tapes = get_completions(save_as_references=False)
    teacher_input_tapes = load_teacher_input_tapes()
    user_input_tapes = load_user_input_tapes()
    teacher_reference_tapes = load_teacher_reference_tapes()
    user_reference_tapes = load_user_reference_tapes()

    # patch envspecs
    for tape in teacher_input_tapes + teacher_reference_tapes:
        assert tape.context
        tape.context.env_spec = forms_path
    for tape in user_input_tapes + user_reference_tapes:
        assert tape.context
        assert tape.context.context
        tape.context.context.env_spec = forms_path

    # with set_sqlite_db_dir(assets_dir):
    teacher_failures = replay_tapes(
        teacher_agent, tapes=teacher_reference_tapes, env=env, start_tapes=teacher_input_tapes, reuse_observations=True
    )
    assert teacher_failures == 0, "Failed to replay teacher tapes"

    user_failures = replay_tapes(
        user_agent, tapes=user_reference_tapes, env=env, start_tapes=user_input_tapes, reuse_observations=True
    )
    assert user_failures == 0, "Failed to replay user tapes"


def test_tape_improver():
    run_dir = f"{res_path}/tape_improver"
    llm = mock_llm(run_dir)
    agent, _, improver_tape = tape_improver.make_world(llm)
    final_tape = tape_improver.CodeImproverTape.model_validate(load_tape_dict(run_dir, "final_tape.json"))
    replay_success = replay_tape(agent, final_tape, start_tape=improver_tape, reuse_observations=True)
    assert replay_success, "Failed to replay tape"


def test_optimize_gpt35():
    assets_dir = f"{res_path}/optimize/gpt-3.5-turbo"
    with run_test_in_tmp_dir(assets_dir):
        with open("config.yaml") as f:
            cfg = DictConfig(yaml.safe_load(f))
        agent = make_agentic_rag_agent(cfg)
        mock_llm = ReplayLLM.from_llm(agent.llms["default"], assets_dir)
        agent.llms = {"default": mock_llm}
        env = make_env()
        tape = DialogTape.model_validate(load_tape_dict(""))
        replay_success = replay_tape(agent, tape, env=env, reuse_observations=True)
        assert replay_success, "Failed to replay tape"


def test_optimize_gpt4o_mini():
    assets_dir = f"{res_path}/optimize/gpt-4o-mini"
    with run_test_in_tmp_dir(assets_dir):
        with open("config.yaml") as f:
            cfg = DictConfig(yaml.safe_load(f))
        agent = make_agentic_rag_agent(cfg)
        mock_llm = ReplayLLM.from_llm(agent.llms["default"], assets_dir)
        agent.llms = {"default": mock_llm}
        env = make_env()
        tape = DialogTape.model_validate(load_tape_dict(""))
        replay_success = replay_tape(agent, tape, env=env, reuse_observations=True)
        assert replay_success, "Failed to replay tape"


def test_optimize_llama33_70b():
    assets_dir = f"{res_path}/optimize/llama-3.3-70b-instruct"
    with run_test_in_tmp_dir(assets_dir):
        with open("config.yaml") as f:
            cfg = DictConfig(yaml.safe_load(f))
        agent = make_agentic_rag_agent(cfg)
        mock_llm = ReplayLLM.from_llm(agent.llms["default"], assets_dir)
        agent.llms = {"default": mock_llm}
        env = make_env()
        tape = DialogTape.model_validate(load_tape_dict(""))
        replay_success = replay_tape(agent, tape, env=env, reuse_observations=True)
        assert replay_success, "Failed to replay tape"


if __name__ == "__main__":
    test_llama_agent()
    test_llama_agent_traces()
    test_llama_agent_tape_reuse()
    test_gaia_agent()
    test_workarena_agent()
    test_delegate()
    test_delegate_stack()
    test_data_science()
    test_form_filler()
    test_tape_improver()
    test_optimize_gpt35()
    test_optimize_gpt4o_mini()
    test_optimize_llama33_70b()



import gzip
import os
import shutil
import sys

_db = "tapedata.sqlite"


def compress_file(file_path):
    with open(file_path, "rb") as f_in:
        with gzip.open(file_path + ".gz", "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    # os.remove(file_path)


def main():
    dirname = sys.argv[1]
    for root, dirs, files in os.walk(dirname):
        for file in files:
            if file.endswith(_db):
                file_path = os.path.join(root, file)
                print(f"Compressing {file_path}")
                compress_file(file_path)


if __name__ == "__main__":
    assert len(sys.argv) == 2, "Usage: python compress_sqlite.py <dirname>"
    main()



"""
I/O routines for Tapes.
"""

import json
import logging
import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Generator, Type

import yaml
from pydantic import TypeAdapter

from tapeagents.agent import Agent
from tapeagents.config import ATTACHMENT_DEFAULT_DIR
from tapeagents.core import Tape, TapeType
from tapeagents.steps import UnknownStep, VideoObservation

logger = logging.getLogger(__name__)


class TapeSaver:
    """A class for saving Tape objects using YAML format.

    This class provides functionality to save Tape objects using a YAML dumper.
    It handles the serialization of Tape objects into YAML format.

    Example:
        ```python
        dumper = yaml.SafeDumper(output_file)
        saver = TapeSaver(dumper)
        saver.save(tape)
        ```
    """

    def __init__(self, yaml_dumper: yaml.SafeDumper):
        """Initialize TapeIOYML with a YAML dumper.

        Args:
            yaml_dumper (yaml.SafeDumper): The YAML dumper instance to use for serialization.
        """
        self._dumper = yaml_dumper

    def save(self, tape: Tape):
        """
        Saves the tape data using the configured dumper.

        Args:
            tape (Tape): The tape object containing the data to be saved.
        """
        self._dumper.represent(tape.model_dump(by_alias=True))


@contextmanager
def stream_yaml_tapes(filename: Path | str, mode: str = "w") -> Generator[TapeSaver, None, None]:
    """Stream YAML tapes to a file.

    This function creates a context manager that allows streaming YAML documents to a file.
    It handles file creation, directory creation if necessary, and proper resource cleanup.

    Args:
        filename (Union[Path, str]): Path to the output YAML file. Can be either a string or Path object.
        mode (str, optional): File opening mode. Defaults to "w" (write mode).

    Yields:
        Generator[TapeSaver, None, None]: A TapeSaver instance that can be used to write YAML documents.

    Raises:
        OSError: If there are issues with file/directory creation or permissions.
        yaml.YAMLError: If there are YAML serialization errors.
    """
    if isinstance(filename, str):
        filename = Path(filename)
    logger.info(f"Writing to {filename} in mode {mode}")

    # Create directory path if it does not exist
    filename.parent.mkdir(parents=True, exist_ok=True)

    # Open file for writing and create dumper instance
    _file = open(filename, mode)
    _dumper = yaml.SafeDumper(
        stream=_file,
        default_flow_style=False,
        explicit_start=True,
        sort_keys=False,
    )
    _dumper.open()

    # Yield the dumper to the caller
    yield TapeSaver(_dumper)

    # Close the dumper and file
    _dumper.close()
    _file.close()


def save_json_tape(tape: Tape, tapes_dir: str, name: str = ""):
    """Save a Tape object to a JSON file.

    Args:
        tape (Tape): The Tape object to be saved
        tapes_dir (str): Directory path where the JSON file will be saved
        name (str, optional): Name of the output JSON file. If empty, tapes_dir is used as the full path.
            If provided without .json extension, it will be added automatically. Defaults to "".

    Example:
        ```python
        tape = Tape(...)
        save_json_tape(tape, "/path/to/dir", "my_tape")
        # Saves to /path/to/dir/my_tape.json
        ```

    """
    if name:
        os.makedirs(tapes_dir, exist_ok=True)
    fname = name if name.endswith(".json") else f"{name}.json"
    fpath = os.path.join(tapes_dir, fname) if name else tapes_dir
    with open(fpath, "w") as f:
        f.write(tape.model_dump_json(indent=4))


def save_tape_images(tape: Tape, images_dir: str):
    os.makedirs(images_dir, exist_ok=True)
    for i, step in enumerate(tape):
        if hasattr(step, "image_path"):
            image_path = os.path.join(images_dir, f"{step.metadata.id}.png")
            shutil.copy(step.image_path, image_path)


def load_tape_dicts(path: Path | str, file_extension: str = ".yaml") -> list[dict]:
    if not os.path.exists(path):
        raise FileNotFoundError(f"File not found: {path}")
    if file_extension not in (".yaml", ".json"):
        raise ValueError(f"Unsupported file extension: {file_extension}")
    if os.path.isdir(path):
        logger.info(f"Loading tapes from dir {path}")
        paths = sorted([os.path.join(path, f) for f in os.listdir(path) if f.endswith(file_extension)])
    else:
        paths = [path]
        file_extension = os.path.splitext(path)[-1]
    tapes = []
    for path in paths:
        with open(path) as f:
            if file_extension == ".yaml":
                data = list(yaml.safe_load_all(f))
            else:
                data = json.load(f)
        if not isinstance(data, list):
            data = [data]
        tapes.extend(data)
    return tapes


def load_tapes(
    tape_class: Type[TapeType],
    path: Path | str,
    file_extension: str = ".yaml",
    attachment_dir: str = None,
) -> list[TapeType]:
    """Load tapes from dir with YAML or JSON files.

    This function loads tapes from a file or directory and converts them into tape objects
    using the specified tape class or type adapter.

    Args:
        tape_class (Type[TapeType]): The class or type adapter used to validate and create tape objects.
        path (Union[Path, str]): Path to a file or directory containing tape configurations.
        file_extension (str, optional): File extension to filter by when loading from directory.
            Must be either '.yaml' or '.json'. Defaults to '.yaml'.
        attachment_dir (str, optional): The directory to use for attachments. If None, a default directory will be used.

    Returns:
        list[TapeType]: A list of validated tape objects.

    Raises:
        FileNotFoundError: If the specified path does not exist.
        ValueError: If an unsupported file extension is provided.

    Example:
        ```python
        tapes = load_tapes(TapeClass, "configs/tapes.yaml")
        tapes = load_tapes(tape_adapter, "configs/tapes", ".json")
        ```
    """
    tapes = []
    data = load_tape_dicts(path, file_extension)
    attachment_dir_resolved = get_attachment_dir(path, attachment_dir)
    for tape_dict in data:
        tape = tape_class.model_validate(tape_dict)
        if attachment_dir_resolved:
            # Update attachment_dir for steps that needs it
            for step in tape:
                if isinstance(step, (VideoObservation)):
                    step.attachment_dir = attachment_dir_resolved
        tapes.append(tape)
    return tapes


def load_legacy_tapes(tape_class: Type[TapeType], path: Path | str, step_class: Type | TypeAdapter) -> list[TapeType]:
    tapes = []
    data = load_tape_dicts(path, ".json")
    for tape_dict in data:
        try:
            tape = tape_class.model_validate(tape_dict)
        except Exception:
            step_dicts = tape_dict["steps"]
            tape_dict["steps"] = []
            tape = tape_class.model_validate(tape_dict)
            step_loader = step_class.model_validate if isinstance(step_class, Type) else step_class.validate_python
            steps = []
            for step_dict in step_dicts:
                try:
                    steps.append(step_loader(step_dict))
                except Exception as e:
                    logger.warning(f"Failed to load step: {e}")
                    steps.append(UnknownStep(content=json.dumps(step_dict, indent=2, ensure_ascii=False)))
            tape.steps = steps
        tapes.append(tape)
    return tapes


def save_agent(agent: Agent, filename: str) -> str:
    with open(filename, "w") as f:
        yaml.dump(agent.model_dump(), f)
    return filename


def get_attachment_dir(tape_path: Path | str, attachment_dir: Path | str) -> str | None:
    """
    Determines the directory to use for tape attachments.

    Args:
        tape_path (Path | str]): Path to a file or directory containing tape configurations.
        attachment_dir (Path | str): The directory to use for attachments. If None, a default directory will be used.

    Returns:
        str | None: The path to the attachment directory if it exists, otherwise None.

    Raises:
        FileNotFoundError: If the provided attachment_dir does not exist.

    Examples:
        >>> get_attachment_dir("data/tapes.yaml", "data/images")
        'data/images'

        >>> get_attachment_dir("data/tapes.yaml", None)
        'data/attachments'

        >>> get_attachment_dir("data/tapes/", None)
        'data/attachments'
    """
    if attachment_dir:
        # Use attachment_dir if provided and exists
        if not Path(attachment_dir).is_dir():
            raise FileNotFoundError(f"Tape attachment directory not found: {attachment_dir}")
        return str(attachment_dir)
    else:
        # Use ATTACHMENT_DEFAULT_DIR if exists
        path_obj = Path(tape_path)
        default_path = path_obj.parent.parent if path_obj.is_file() else path_obj.parent
        default_path /= ATTACHMENT_DEFAULT_DIR
        if Path(default_path).is_dir():
            logger.info(f"Use tape attachment director: {default_path}")
            return str(default_path)
    return None



"""Module to optimize agents"""

import logging
import random
from typing import Callable

from pydantic import BaseModel

from .agent import Agent
from .dialog_tape import Tape
from .llm_function import LLMFunctionNode

logger = logging.getLogger(__name__)


class OptimizationResult(BaseModel):
    id: int
    agent: Agent
    metric: float


def optimize_demos(
    agent: Agent,
    good_tapes: list[Tape],
    n_demos: int,
    n_iterations: int,
    seed: int,
    metric_fn: Callable[[list[Tape], str], float],
    run_agent_fn: Callable[[Agent, list[Tape]], list[Tape]],
    post_run_agent_fn: Callable[[OptimizationResult], None] | None,
) -> OptimizationResult:
    """
    Iteratively adds demonstration tapes to the agent, runs the agent, and evaluates its performance
    to find the optimal set of demonstrations that maximize the agent's performance metric.

    Parameters:
    agent (Agent): The initial agent to be optimized.
    good_tapes (list[Tape]): A list of good demonstration tapes.
    n_demos (int): The number of demonstrations to add to the agent.
    n_iterations (int): The number of tries to select the best set of demoinstration.
    seed (int): The initial seed for random number generation, incremented in each iteration to ensure variability per iteration.
    metric_fn (Callable[[list[Tape]], float]): A function to compute the performance metric from a list of tapes.
    run_agent_fn (Callable[[Agent, list[Tape]], list[Tape]]): A function to run the agent and return the resulting tapes.
    post_run_agent_fn (Callable[[OptimizationResult], None]): A callback function to be called after each iteration.

    Returns:
    OptimizationResult: Agents and metrics for each iterations, including the best performing one.
    """
    best_metric = 0
    best_agent = agent
    best_agent_id = 0

    for i in range(n_iterations):
        logger.debug(f"Optimization iteration {i}")
        # Add demos to the agent with a different seed for each attempt
        new_agent = add_demos(agent, good_tapes, n_demos, seed=seed + i)
        # Run agent on the validation set to get metric to optimize
        final_tapes = run_agent_fn(new_agent)
        metric = metric_fn(final_tapes, f"optimization_{i}")
        # Update best agent if necessary
        if metric > best_metric:
            best_metric = metric
            best_agent = new_agent
            best_agent_id = i
        # Callback results
        if post_run_agent_fn:
            post_run_agent_fn(OptimizationResult(id=i, agent=new_agent, metric=metric))
    # Return the best agent
    return OptimizationResult(id=best_agent_id, agent=best_agent, metric=best_metric)


def add_demos(agent: Agent, tapes: list[Tape], max_n_demos: int, seed: int = 1) -> Agent:
    """
    Extract demos for function templates from the given tapes.

    When there is too many demos, select random ones.
    """
    demos = {template_name: [] for template_name in agent.templates}
    for tape in tapes:
        for node, index in agent.get_node_runs(tape):
            if isinstance(node, LLMFunctionNode):
                demos[node.template_name].append(node.extract_demo(agent, tape, index))
    rng = random.Random(seed)
    agent_copy = agent.model_copy(deep=True)
    for template_name, template in agent_copy.templates.items():
        k = min(max_n_demos, len(demos[template_name]))
        template.demos = rng.sample(demos[template_name], k)
    return agent_copy



import logging
import os
import sys
from collections import defaultdict

import uvicorn
from pydantic import TypeAdapter

from tapeagents.core import Action
from tapeagents.io import load_legacy_tapes
from tapeagents.observe import retrieve_all_llm_calls
from tapeagents.renderers.camera_ready_renderer import CameraReadyRenderer
from tapeagents.tape_browser import TapeBrowser

from ..eval import calculate_accuracy, get_exp_config_dict, load_dataset, tape_correct
from ..steps import GaiaStep, GaiaTape

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger(__name__)


class GaiaTapeBrowser(TapeBrowser):
    def __init__(self, tapes_folder: str, renderer):
        super().__init__(tape_cls=GaiaTape, tapes_folder=tapes_folder, renderer=renderer, file_extension=".json")

    def load_tapes(self, name: str) -> list:
        _, exp_dir, postfix = name.split("/", maxsplit=2)
        try:
            cfg_dir = os.path.join(self.tapes_folder, exp_dir)
            cfg = get_exp_config_dict(cfg_dir)
            try:
                tasks = load_dataset(cfg["split"])
                tasks_list = [task for level in tasks.values() for task in level]
                self.task_id_to_num = {task["task_id"]: i + 1 for i, task in enumerate(tasks_list)}
            except Exception:
                self.task_id_to_num = {}
        except Exception as e:
            logger.exception(f"Failed to load config from {cfg_dir}: {e}")
            self.task_id_to_num = {}
        tapes_path = os.path.join(self.tapes_folder, exp_dir, "tapes")
        image_dir = os.path.join(self.tapes_folder, exp_dir, "attachments", "images")
        if not os.path.exists(image_dir):
            image_dir = os.path.join(self.tapes_folder, exp_dir, "images")

        try:
            all_tapes: list[GaiaTape] = load_legacy_tapes(GaiaTape, tapes_path, step_class=TypeAdapter(GaiaStep))  # type: ignore
        except Exception as e:
            logger.error(f"Failed to load tapes from {tapes_path}: {e}")
            return []
        tapes = []
        for tape in all_tapes:
            if postfix == "all" or str(tape.metadata.level) == postfix:
                tapes.append(tape)
            for i in range(len(tape.steps)):
                if tape.steps[i].kind != "image":
                    continue
                image_path = os.path.join(image_dir, f"{tape.steps[i].metadata.id}.png")
                if os.path.exists(image_path):
                    tape.steps[i].metadata.other["image_path"] = os.path.join(
                        exp_dir, "images", f"{tape.steps[i].metadata.id}.png"
                    )
                else:
                    logger.warning(f"Image not found: {image_path}")

        self.llm_calls = {}
        sqlite_fpath = os.path.join(self.tapes_folder, exp_dir, "tapedata.sqlite")
        if not os.path.exists(sqlite_fpath):
            sqlite_fpath = os.path.join(self.tapes_folder, exp_dir, "llm_calls.sqlite")
        try:
            llm_calls = retrieve_all_llm_calls(sqlite_fpath)
            self.llm_calls = {llm_call.prompt.id: llm_call for llm_call in llm_calls}
        except Exception as e:
            logger.error(f"Failed to load LLM calls from {sqlite_fpath}: {e}")
        logger.info(f"Loaded {len(tapes)} tapes from {tapes_path}")
        logger.info(f"Loaded {len(self.llm_calls)} prompts from {sqlite_fpath}")
        return tapes

    def load_llm_calls(self):
        pass

    def update_tape_view(self, tape_id: int) -> tuple[str, str]:
        try:
            tape: GaiaTape = self.tapes[tape_id]  # type: ignore
        except IndexError:
            return "", "Tape not found"
        label = self.get_tape_label(tape)
        html = f"{self.renderer.style}<style>.thought {{ background-color: #ffffba !important; }};</style>{self.renderer.render_tape(tape, self.llm_calls)}"
        return html, label

    def get_exp_label(self, filename: str, tapes: list[GaiaTape]) -> str:
        acc, n_solved = calculate_accuracy(tapes)
        errors = defaultdict(int)
        prompt_tokens_num = 0
        output_tokens_num = 0
        total_cost = 0.0
        visible_prompt_tokens_num = 0
        visible_output_tokens_num = 0
        visible_cost = 0.0
        no_result = 0
        actions = defaultdict(int)
        for llm_call in self.llm_calls.values():
            prompt_tokens_num += llm_call.prompt_length_tokens
            output_tokens_num += llm_call.output_length_tokens
            total_cost += llm_call.cost
        for tape in tapes:
            if tape.metadata.result in ["", None, "None"]:
                no_result += 1
            if tape.metadata.error:
                errors["fatal"] += 1
            if tape.metadata.terminated:
                errors["terminated"] += 1
            last_action = None
            counted = set([])
            for step in tape:
                llm_call = self.llm_calls.get(step.metadata.prompt_id)
                if llm_call and step.metadata.prompt_id not in counted:
                    counted.add(step.metadata.prompt_id)
                    visible_prompt_tokens_num += llm_call.prompt_length_tokens
                    visible_output_tokens_num += llm_call.output_length_tokens
                    visible_cost += llm_call.cost
                if isinstance(step, Action):
                    actions[step.kind] += 1
                    last_action = step
                if step.kind == "search_results_observation" and not step.serp:
                    errors["search_empty"] += 1
                if step.kind == "page_observation" and step.error:
                    errors["browser"] += 1
                elif step.kind == "llm_output_parsing_failure_action":
                    errors["parsing"] += 1
                elif step.kind == "action_execution_failure":
                    if last_action:
                        errors[f"{last_action.kind}"] += 1
                    else:
                        errors["unknown_action_execution_failure"] += 1
                elif step.kind == "code_execution_result" and step.result.exit_code:
                    errors["code_execution"] += 1
        timers, timer_counts = self.aggregate_timer_times(tapes)
        html = f"<h2>Solved {acc:.2f}%, {n_solved} out of {len(tapes)}</h2>"
        if "all" in filename:
            html += f"Prompt tokens: {prompt_tokens_num}<br>Output tokens: {output_tokens_num}<br>Cost: {total_cost:.2f} USD<h3>Visible</h3>"
        html += f"Prompt tokens: {visible_prompt_tokens_num}<br>Output tokens: {visible_output_tokens_num}<br>Cost: {visible_cost:.2f} USD"
        if errors:
            errors_str = "<br>".join(f"{k}: {v}" for k, v in errors.items())
            html += f"<h2>No result: {no_result}</h2>"
            html += f"<h2>Errors: {sum(errors.values())}</h2>{errors_str}"
        if actions:
            actions_str = "<br>".join(f"{k}: {v}" for k, v in actions.items())
            html += f"<h2>Actions: {sum(actions.values())}</h2>{actions_str}"
        if timers:
            timers_str = "<br>".join(
                f"{'execute ' if k.endswith('action') else ''}{k}: {v:.1f} sec, avg. {v/timer_counts[k]:.1f} sec"
                for k, v in timers.items()
            )
            html += f"<h2>Timings</h2>{timers_str}"
        return html

    def get_tape_name(self, i: int, tape: GaiaTape) -> str:
        error = "F" if tape.metadata.error else ""
        if tape.metadata.terminated:
            error = "T"
        last_action = None
        tokens = 0
        for step in tape:
            llm_call = self.llm_calls.get(step.metadata.prompt_id)
            tokens += llm_call.prompt_length_tokens if llm_call else 0
            tokens += llm_call.output_length_tokens if llm_call else 0
            if isinstance(step, Action):
                last_action = step
            if step.kind == "search_results_observation" and not step.serp:
                error += "se"
            elif step.kind == "page_observation" and step.error:
                error += "br"
            elif step.kind == "llm_output_parsing_failure_action":
                error += "pa"
            elif step.kind == "action_execution_failure" and last_action:
                error += last_action.kind[:2]
            elif step.kind == "code_execution_result" and step.result.exit_code:
                error += "ce"
        mark = "+" if tape_correct(tape) else ("" if tape.metadata.result else "∅")
        if tape.metadata.task.get("file_name"):
            mark += "📁"
        if error:
            mark += f"[{error}]"
        if mark:
            mark += " "
        n = self.task_id_to_num.get(tape.metadata.task.get("task_id"), "")
        name = tape[0].content[:32] if hasattr(tape[0], "content") else tape[0].short_view()[:32]
        return f"{n} {mark}({tokens: }t) {name}"  # type: ignore

    def get_tape_label(self, tape: GaiaTape) -> str:
        llm_calls_num = 0
        input_tokens_num = 0
        output_tokens_num = 0
        cost = 0

        for step in tape:
            prompt_id = step.metadata.prompt_id
            if prompt_id and prompt_id in self.llm_calls:
                llm_calls_num += 1
                if prompt_id in self.llm_calls:
                    llm_call = self.llm_calls[prompt_id]
                    input_tokens_num += llm_call.prompt_length_tokens
                    output_tokens_num += llm_call.output_length_tokens
                    cost += llm_call.cost
        failure_count = len(
            [step for step in tape if "failure" in step.kind or (step.kind == "page_observation" and step.error)]
        )
        success = tape[-1].success if hasattr(tape[-1], "success") else ""  # type: ignore
        overview = tape[-1].overview if hasattr(tape[-1], "overview") else ""  # type: ignore
        label = f"""<h2>Result</h2>
            <div class="result-label expected">Golden Answer: <b>{tape.metadata.task.get('Final answer', '')}</b></div>
            <div class="result-label">Agent Answer: <b>{tape.metadata.result}</b></div>
            <div class="result-success">Finished successfully: {success}</div>
            <h2>Summary</h2>
            <div class="result-overview">{overview}</div>
            <h2>Stats</h2>
            <div class="result-label">Steps: {len(tape)}</div>
            <div class="result-label">Failures: {failure_count}</div>
            <div>LLM Calls: {llm_calls_num}</div>
            <div>Input tokens: {input_tokens_num}</div>
            <div>Output tokens: {output_tokens_num}</div>
            <div>Cost: {cost:.2f} USD</div>
        """
        if tape.metadata.error:
            label += f"<div class='result-error'><b>Error</b>: {tape.metadata.error}</div>"
        return label

    def get_tape_files(self) -> list[str]:
        raw_exps = [
            d for d in os.listdir(self.tapes_folder) if os.path.isdir(os.path.join(self.tapes_folder, d, "tapes"))
        ]
        assert raw_exps, f"No experiments found in {self.tapes_folder}"
        logger.info(f"Found {len(raw_exps)} experiments")
        exps = []
        for postfix in ["1", "2", "3", "all"]:
            for r in raw_exps:
                exp_dir = os.path.join(self.tapes_folder, r)
                try:
                    try:
                        cfg = get_exp_config_dict(exp_dir)
                    except Exception:
                        cfg = {}
                    if "split" in cfg:
                        set_name = cfg["split"]
                    elif "data_dir" in cfg:
                        parts = cfg["data_dir"].split("/")
                        set_name = parts[-2] if cfg["data_dir"].endswith("/") else parts[-1]
                    else:
                        set_name = ""
                except Exception as e:
                    logger.error(f"Failed to load config from {exp_dir}: {e}")
                    set_name = ""
                exps.append(f"{set_name}/{r}/{postfix}")
        return sorted(exps)

    def aggregate_timer_times(self, tapes: list[GaiaTape]):
        timer_sums = defaultdict(float)
        timer_counts = defaultdict(int)
        for tape in tapes:
            timers = tape.metadata.other.get("timers", {})
            for timer_name, exec_time in timers.items():
                timer_sums[timer_name] += exec_time
                timer_counts[timer_name] += 1
            for step in tape.steps:
                action_kind = step.metadata.other.get("action_kind")
                action_execution_time = step.metadata.other.get("action_execution_time")
                if action_kind and action_execution_time:
                    timer_sums[action_kind] += action_execution_time
                    timer_counts[action_kind] += 1
        return dict(timer_sums), dict(timer_counts)


def main(dirname: str):
    browser = GaiaTapeBrowser(dirname, CameraReadyRenderer())
    browser.launch(port=7861)


uvicorn.config.LOOP_SETUPS = {  # type: ignore
    "none": None,
    "auto": "uvicorn.loops.asyncio:asyncio_setup",
    "asyncio": "uvicorn.loops.asyncio:asyncio_setup",
    "uvloop": "uvicorn.loops.uvloop:uvloop_setup",
}

if __name__ == "__main__":
    assert len(sys.argv) == 2, "Usage: python -m scripts.tape_browser <dirname>"
    main(sys.argv[1])



import os
import sys
from pathlib import Path

from tapeagents.renderers.camera_ready_renderer import CameraReadyRenderer
from tapeagents.tape_browser import TapeBrowser
from tapeagents.team import TeamTape

tape_dir = Path(sys.argv[1])
os.environ["TAPEAGENTS_SQLITE_DB"] = os.path.join(tape_dir, "tapedata.sqlite")


browser = TapeBrowser(TeamTape, sys.argv[1], CameraReadyRenderer(), file_extension=".json")
browser.launch(port=7680 if len(sys.argv) < 3 else int(sys.argv[2]))



import logging
import sys

from tapeagents.dialog_tape import DialogTape, SystemStep, UserStep
from tapeagents.llms import LLM, TrainableLLM
from tapeagents.renderers.pretty import PrettyRenderer
from tapeagents.studio import Studio

from .delegate_stack import EXAMPLE_TEXT, ExampleTape, make_analyze_text_chain
from .llama_agent import LLAMAChatBot

logging.basicConfig(level=logging.INFO)


def try_studio_with_stack(llm: LLM):
    """
    Launches the studio with the stack of agents that analyze the text for nouns and irregular verbs.
    """
    tape = ExampleTape(context=EXAMPLE_TEXT)
    agent = make_analyze_text_chain(llm)
    Studio(agent, tape, PrettyRenderer()).launch()


def try_studio_with_chat(llm: LLM):
    """
    Launches the studio with the agent that responds to the user using the style of Shakespeare books.
    """
    tape = DialogTape(
        context=None,
        steps=[
            SystemStep(
                content="Respond to the user using the style of Shakespeare books. Be very brief, 50 words max."
            ),
            UserStep(content="Hello, how are you?"),
        ],
    )
    agent = LLAMAChatBot.create(llm)
    Studio(agent, tape, PrettyRenderer()).launch()


# Interactive Gradio demo of the agent that could be changed in runtime.
if __name__ == "__main__":
    llm = TrainableLLM(
        base_url="https://api.together.xyz",
        model_name="meta-llama/Meta-Llama-3-70B-Instruct-Turbo",
        parameters=dict(temperature=0.7, max_tokens=512),
    )
    if len(sys.argv) < 2:
        try_studio_with_chat(llm)
    elif sys.argv[1] == "chat":
        try_studio_with_chat(llm)
    elif sys.argv[1] == "stack":
        try_studio_with_stack(llm)
    else:
        raise ValueError(f"Unknown mode, {sys.argv[1]}")



import logging
import os

import gradio as gr
import hydra
import nest_asyncio
import uvicorn
from browsergym.workarena import ALL_WORKARENA_TASKS
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from omegaconf import DictConfig

from tapeagents.llms import LLM
from tapeagents.studio import Studio

from .agent import WorkArenaAgent, WorkArenaBaseline
from .environment import WorkArenaEnvironment
from .tape_browser import WorkArenaRender

nest_asyncio.apply()

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

logger = logging.getLogger(__name__)


@hydra.main(
    version_base=None,
    config_path="../../conf",
    config_name="workarena_demo",
)
def main(cfg: DictConfig) -> None:
    os.environ["TAPEAGENTS_SQLITE_DB"] = os.path.join(cfg.exp_path, "tapedata.sqlite")
    llm: LLM = hydra.utils.instantiate(cfg.llm)
    env = WorkArenaEnvironment(**cfg.env)
    if cfg.agent == "baseline":
        agent = WorkArenaBaseline.create(llm)
    else:
        logger.info("Use guided agent")
        agent = WorkArenaAgent.create(llm)
    tape, _ = env.start_task(ALL_WORKARENA_TASKS[0], seed=cfg.seeds[0])
    blocks = Studio(agent, tape, WorkArenaRender(exp_dir=""), env).blocks
    logger.info(f"Starting FastAPI server with static dir {cfg.exp_path}")
    app = FastAPI()
    app.mount("/static", StaticFiles(directory=cfg.exp_path), name="static")
    app = gr.mount_gradio_app(app, blocks, path="/")
    uvicorn.run(app, host="0.0.0.0", port=7860)


if __name__ == "__main__":
    main()



import json
import logging
from typing import Any, Generator

import pydantic

from tapeagents.agent import Agent, Node
from tapeagents.core import LLMOutputParsingFailureAction, Prompt, SetNextNode
from tapeagents.dialog_tape import AssistantStep
from tapeagents.llms import LLM, LLMStream
from tapeagents.utils import sanitize_json_completion

from .error import FormFillerStateError
from .state import compute_form_filler_state, update_form_filler_state
from .steps import (
    AnswerFromFunctionSchema,
    CallFunction,
    Exit,
    FormFillerStep,
    InspectFunction,
    NoAnswerFromFunctionSchema,
    RefuseInexistentFunction,
    RefuseInvalidFunctionParameterSkip,
    RefuseInvalidFunctionParameterValue,
    RefuseToEngage,
    RequestExitConfirmation,
    RequestFunction,
    RequestFunctionCallConfirmation,
    RequestFunctionParameters,
    ResolveFunction,
    UpdateFunctionParameters,
)
from .tape import FormFillerTape, prepare_formfiller_template_variables
from .types import FunctionName

logger = logging.getLogger(__name__)


class StudentAgent(Agent[FormFillerTape]):
    @classmethod
    def create(cls, llm: LLM, templates: dict[str, dict[str, str]]):
        nodes = [
            CheckFunctionCandidatesNode(),
            IntentDiscoveryNode(),
            StudentNode(),
        ]
        return super().create(name="student_formfiller", llms=llm, templates=templates, nodes=nodes)

    def generate_steps(self, tape: FormFillerTape, llm_stream: LLMStream) -> Generator[FormFillerStep, None, None]:
        # compute current tape state
        state = compute_form_filler_state(tape)
        assert not isinstance(
            state, FormFillerStateError
        ), f"Tape state should not be an error.\nTAPE:\n{tape}\nSTATE:\n{state}"
        # try to update state for all predicted steps
        for predicted_step in super().generate_steps(tape, llm_stream):
            assert isinstance(
                predicted_step, FormFillerStep
            ), f"Predicted step should be a FormFillerStep, got {predicted_step}"
            state_or_error = update_form_filler_state(state, predicted_step)
            # stop as soon as there is an error
            if isinstance(state_or_error, FormFillerStateError):
                yield state_or_error  # type: ignore
                return
            # yield the step if state updated successfully
            state = state_or_error
            yield predicted_step


class CheckFunctionCandidatesNode(Node):
    name: str = "check_function_candidates_node"

    def generate_steps(
        self, agent: StudentAgent, tape: FormFillerTape, llm_stream: LLMStream
    ) -> Generator[FormFillerStep, None, None]:
        if not tape.are_function_candidates_listed:
            assert tape.last_user_step is not None, "User step is required to resolve function"
            yield ResolveFunction(query=tape.last_user_step.content)
        else:
            yield SetNextNode(next_node="intent_discovery_node")


class IntentDiscoveryNode(Node):
    name: str = "intent_discovery_node"

    def make_prompt(self, agent: StudentAgent, tape: FormFillerTape) -> Prompt:
        # If no function has been inspected
        if not tape.intent_is_discovered:
            template_variables = self.create_template_variables(tape)
            return Prompt.from_user_message(
                agent.templates["intent_discovery_post_list_template"].format(**template_variables)
            )
        else:
            return Prompt()

    def create_template_variables(self, tape: FormFillerTape) -> dict[str, Any]:
        return {
            "function_search_results": json.dumps(tape.function_candidates_step.model_dump()["candidates"], indent=2)
            if tape.function_candidates_step
            else "< Not available >",
            "last_user_message": tape.last_user_step.content if tape.last_user_step else "< Not available >",
        }

    def generate_steps(
        self, agent: StudentAgent, tape: FormFillerTape, llm_stream: LLMStream
    ) -> Generator[FormFillerStep, None, None]:
        if llm_stream:
            inspect_function_found = False
            predicted_steps = parse_completion(None, llm_stream.get_text())
            for step in predicted_steps:
                if isinstance(step, InspectFunction):
                    inspect_function_found = True
                yield step

            if not inspect_function_found:
                yield SetNextNode(next_node="intent_discovery_node")
        else:
            yield SetNextNode(next_node="student_node")


class StudentNode(Node):
    name: str = "student_node"

    def make_prompt(self, agent: StudentAgent, tape: FormFillerTape) -> Prompt:
        template_variables = prepare_formfiller_template_variables(tape)
        return Prompt.from_user_message(agent.templates["main_template"].format(**template_variables))

    def generate_steps(
        self, agent: StudentAgent, tape: FormFillerTape, llm_stream: LLMStream
    ) -> Generator[FormFillerStep, None, None]:
        assert tape.function_schema_step is not None, "Function schema should already be determined in student node"
        yield from parse_completion(tape.function_schema_step.name, llm_stream.get_text())
        # go back to this node if the agent gets called again
        yield SetNextNode(next_node="student_node")


def parse_completion(function_name: FunctionName | None, completion: str) -> Generator[FormFillerStep, None, None]:
    # first, check if the completion has a valid json format
    try:
        step_dict = json.loads(sanitize_json_completion(completion))
    except Exception as e:
        logger.exception(f"Failed to parse agent output: {completion}\n\nError: {e}")
        yield LLMOutputParsingFailureAction(error=f"Failed to parse agent output.\n\nError: {e}", llm_output=completion)
        return
    try:
        yield from reconstruct_compact_steps(function_name, step_dict)
    except pydantic.ValidationError as e:
        logger.exception(f"Failed to parse step data: {step_dict}\n\nError: {e}")
        yield LLMOutputParsingFailureAction(error=f"Failed to parse agent output.\n\nError: {e}", llm_output=completion)


def reconstruct_compact_steps(
    function_name: FunctionName | None, data: dict[str, Any]
) -> Generator[FormFillerStep, None, None]:
    data.setdefault("thoughts", [])

    for entry in data["thoughts"]:
        match entry:
            # Intent classification
            case "no_form_found":
                yield RefuseInexistentFunction()
            case {"offer_forms": functions}:
                yield RequestFunction(functions=functions)

            # question answering
            case "answer":
                assert isinstance(function_name, str), f"Function must be selected, {data}"
                yield AnswerFromFunctionSchema(function=function_name)
            case "no_answer":
                assert isinstance(function_name, str), f"Function must be selected, {data}"
                yield NoAnswerFromFunctionSchema(function=function_name)

            # unrelated requests
            case "unrelated_request":
                yield RefuseToEngage()

            # Update function parameters"
            case {"update_parameters": updates}:
                assert isinstance(function_name, str), f"Function must be selected, {data}"
                assigns = {}
                skips = []
                for key, value in updates.items():
                    if value == "_skip":
                        skips.append(key)
                    elif value == "_refuse_value":
                        yield RefuseInvalidFunctionParameterValue(
                            function=function_name,
                            parameter=key,
                            parameter_value="unk",
                        )
                    elif value == "_refuse_skip":
                        yield RefuseInvalidFunctionParameterSkip(function=function_name, parameter=key)
                    else:
                        assigns[key] = value
                if assigns or skips:
                    yield UpdateFunctionParameters(function=function_name, assign=assigns, skip=skips)

            # Confirm or request
            case "confirm_exit":
                yield RequestExitConfirmation()
            case "confirm_submit":
                assert isinstance(function_name, str), f"Function must be selected, {data}"
                yield RequestFunctionCallConfirmation(function=function_name)
            case {"next_requested_parameter": parameter}:
                assert isinstance(function_name, str), f"Function must be selected, {data}"
                yield RequestFunctionParameters(function=function_name, parameters=[parameter])
            case _:
                yield LLMOutputParsingFailureAction(error=f"Unexpected entry: {entry}", llm_output=str(data))
                return

    if "action" not in data:
        yield LLMOutputParsingFailureAction(error="No action found in the compact steps", llm_output=str(data))
        return

    match data["action"]:
        case "list_forms":
            yield ResolveFunction(query=None)
        case {"select_form": function_name}:
            assert isinstance(function_name, str), f"Function name must be a string in {data}"
            yield InspectFunction(function=function_name)
        case "exit":
            yield Exit()
        case "submit":
            assert isinstance(function_name, str), "Function name is required for submit"
            yield CallFunction(function=function_name)
        case {"prompt": message}:
            yield AssistantStep(content=message)
        case _:
            yield LLMOutputParsingFailureAction(error=f"Unexpected action: {data['action']}", llm_output=str(data))



import logging
import math
import os
import re
import time
from pathlib import Path
from typing import Optional

import ffmpeg
import webvtt
import whisper
import yt_dlp
from whisper.utils import get_writer

from tapeagents.config import ATTACHMENT_DEFAULT_DIR
from tapeagents.core import Action, Observation
from tapeagents.steps import VideoObservation, WatchVideoAction
from tapeagents.tools.base import Tool

logger = logging.getLogger(__name__)


def get_video_observation(
    url: str,
    output_dir: str,
    start_time: str = "",
    end_time: str = "",
) -> VideoObservation:
    try:
        video_path, thumbnail_path = download_video(url, output_dir)
        video_path_trimmed = trim_video(video_path, start_time, end_time)
        video_contact_sheet_paths = generate_contact_sheets_from_video(
            video_path, video_path_trimmed=video_path_trimmed, start_time=start_time, end_time=end_time
        )
        try:
            subtitle_path = transcribe_audio(video_path, video_path_trimmed, start_time=start_time, end_time=end_time)
        except Exception as e:
            logger.exception(f"Error while transcribing audio: {e}")
            subtitle_path = None
        try:
            subtitle_text = extract_text_from_vtt(subtitle_path, start_time, end_time)
        except Exception as e:
            logger.exception(f"Error while extracting text from VTT: {e}")
            subtitle_text = None
        error = None
    except Exception as e:
        logger.exception(f"Error while watching video: {e}")
        raise e
    video_observation = VideoObservation(
        attachment_dir=output_dir,
        video_path=Path(video_path_trimmed).relative_to(output_dir).as_posix(),
        video_contact_sheet_paths=[Path(path).relative_to(output_dir).as_posix() for path in video_contact_sheet_paths],
        thumbnail_path=Path(thumbnail_path).relative_to(output_dir).as_posix(),
        subtitle_path=Path(subtitle_path).relative_to(output_dir).as_posix(),
        subtitle_text=subtitle_text,
        error=error,
    )
    return video_observation


def download_video(url: str, output_dir: str) -> str:
    if "youtube" in url:
        video = download_video_youtube(url, output_dir)
        time.sleep(2)
        return video
    else:
        raise NotImplementedError("Only youtube videos are supported at the moment")


def trim_video(video_path: str, start_time: str, end_time: str) -> str:
    if not video_path:
        raise ValueError("video_path is required")

    if not start_time and not end_time:
        return video_path

    if not start_time:
        start_time = "00:00:00"

    if not end_time:
        probe = ffmpeg.probe(video_path)
        video_duration = float(probe["streams"][0]["duration"])
        end_time = str(video_duration)

    if start_time == end_time:
        start_time = str(float(end_time) - 0.5)
        end_time = str(float(end_time) + 0.5)

    try:
        logger.info("Trimming video")
        video_base_path, ext = os.path.splitext(video_path)
        trimmed_video_path = f"{video_base_path}_{start_time.replace(':', '-')}_{end_time.replace(':', '-')}{ext}"
        ffmpeg.input(video_path).output(
            trimmed_video_path,
            ss=start_time,
            to=end_time,
        ).run(overwrite_output=True)
    except ffmpeg.Error as e:
        raise e
    return trimmed_video_path


def download_video_youtube(url: str, output_dir: str) -> str:
    ydl_opts = {
        "format": "best[ext=mp4]/bestvideo[ext=mp4]+bestaudio[ext=m4a]/best",  # Download best mp4 format available or any other best if no mp4 available
        "formatsort": "res:360",  # Video available with the largest resolution but no better than 360p
        "writethumbnail": True,
        "writesubtitles": True,  # Official docs mention write-subs
        "writeautomaticsub": True,  # Official docs mention write-auto-subs
        "paths": {"home": output_dir},
        "outtmpl": "%(id)s.%(ext)s",  # Output file name template
        "logger": YTDLogger(),
        "progress_hooks": [ytd_progress_hook],
    }

    video_id = get_video_id(url)
    video_path = find_file(output_dir, video_id, (".mp4", ".webm"))
    thumbnail_path = find_file(output_dir, video_id, (".webp", ".jpg", ".jpeg", ".png"), strict=True)
    if not video_path:  # video not in cache
        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            try:
                ydl.download(url)
            except yt_dlp.utils.YoutubeDLError as e:
                # Output list format for debug
                ydl_opts["listformats"] = True
                with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                    ydl.download(url)
                raise e
        video_path = find_file(output_dir, video_id, (".mp4", ".webm"))
        thumbnail_path = find_file(output_dir, video_id, (".webp", ".jpg", ".jpeg", ".png"), strict=True)
    return video_path, thumbnail_path


def get_video_id(url: str) -> str:
    """
    Extract the video ID from a YouTube URL.
    Supports various YouTube URL formats.
    """
    video_id = None
    patterns = [
        r"v=([^&]+)",  # e.g. https://www.youtube.com/watch?v=VIDEO_ID
        r"youtu\.be/([^?&]+)",  # e.g. https://youtu.be/VIDEO_ID
        r"youtube\.com/embed/([^?&]+)",  # e.g. https://www.youtube.com/embed/VIDEO_ID
    ]
    for pattern in patterns:
        match = re.search(pattern, url)
        if match:
            video_id = match.group(1)
            break
    if not video_id:
        raise ValueError("Invalid YouTube URL")
    return video_id


def find_file(dir: str, filename: str, file_extentions: tuple, strict: bool = False) -> Optional[str]:
    if strict:
        for file_extention in file_extentions:
            file = f"{filename}{file_extention}"
            file_path = os.path.join(dir, file)
            if os.path.exists(file_path):
                return file_path
        return None

    for file in os.listdir(dir):
        if file.startswith(filename) and file.endswith(file_extentions):
            return os.path.join(dir, file)
    return None


def generate_contact_sheets_from_video(
    video_path: str,
    video_path_trimmed: str,
    max_contact_sheets: int = 1,
    frame_interval_seconds: int = 0,  # 0 means auto frame interval
    start_time: str = "",
    end_time: str = "",
) -> list[str]:
    # Optimized grid for 512px x 512px: https://platform.openai.com/docs/guides/vision
    nb_tile_x = 3
    nb_tile_y = 5
    scale = "320:180"
    format = "png"
    output_suffix = "_contact-sheet_"

    # Return existing contact sheets if they exist
    output_paths = []
    output_dir = os.path.dirname(video_path_trimmed)
    file_base_path, _ = os.path.splitext(os.path.basename(video_path_trimmed))
    for file in os.listdir(output_dir):
        if re.search(r"{}{}\d+".format(file_base_path, output_suffix), file):
            output_paths.append(os.path.join(output_dir, file))
    if len(output_paths) > 0:
        logger.info(f"Use cached contact sheets: {output_paths}")
        return output_paths

    # Determine the number of frames and contact sheets needed
    try:
        probe = ffmpeg.probe(video_path)
    except ffmpeg.Error as e:
        logger.error(f"ffprobe error: {e.stderr.decode('utf-8')}")
        raise e
    video_duration = float(probe["streams"][0]["duration"])  # Video duration in seconds
    fps = eval(probe["streams"][0]["r_frame_rate"])  # Extract FPS

    start_time_seconds = time_to_seconds(start_time) if start_time else 0
    end_time_seconds = time_to_seconds(end_time) if end_time else video_duration
    end_time_seconds = min(end_time_seconds, video_duration)

    duration = end_time_seconds - start_time_seconds
    if frame_interval_seconds == 0:  # Auto frame interval
        frame_interval_seconds = duration / (max_contact_sheets * nb_tile_x * nb_tile_y)
    elif duration <= frame_interval_seconds:  # If the video is shorter than the frame interval, use the whole video
        frame_interval_seconds = duration
    logger.info(f"Frame interval (s): {frame_interval_seconds}")
    total_nb_frames = int(duration / frame_interval_seconds)
    nb_tile_y = min(nb_tile_y, math.ceil(total_nb_frames / nb_tile_x))
    nb_tile_x = min(nb_tile_x, math.ceil(total_nb_frames / nb_tile_y))
    frames_per_contact_sheet = nb_tile_x * nb_tile_y
    total_contact_sheets = math.ceil(total_nb_frames / frames_per_contact_sheet)
    frames_per_interval = int(fps * frame_interval_seconds)

    # Generate contact sheets
    vf = """drawtext=text='%{pts\\:hms}'
            :x='(main_w-text_w)/2'
            :y='(main_h-text_h)'
            :fontcolor='Yellow'
            :fontsize='(main_h/12)'
            :boxcolor='Black':box=1
            """  # Add a timestamp to each frame
    vf += f",thumbnail=n={frames_per_interval}"  # Pick one of the most representative frames in sequences of nb_frames consecutive frames
    vf += f",scale={scale}"  # Resize the frames to a specific size
    vf += ",pad=iw+4:ih+4:2:2:color=black"  # Add borders around each frame
    vf += f",tile={nb_tile_x}x{nb_tile_y}"  # Arrange the frames in a grid

    logger.info("Generating contact sheet from video")
    output_paths = []
    video_base_path, _ = os.path.splitext(video_path_trimmed)
    for i in range(total_contact_sheets):
        start_frame = i * frame_interval_seconds * frames_per_contact_sheet
        ss = start_time_seconds + start_frame
        to = ss + duration
        output_path = f"{video_base_path}{output_suffix}{i+1}.{format}"
        try:
            cmd = ffmpeg.input(video_path, to=to).output(
                output_path,
                ss=ss,
                vframes=1,
                vf=vf,
            )
            logger.info(f"FFMPEG ARGS: {cmd.get_args()}")
            cmd.run(overwrite_output=True)
            output_paths.append(output_path)
        except ffmpeg.Error as e:
            logger.error(f"ffmpeg error: {e.stderr.decode('utf-8')}")
            raise e
    logger.info(f"Generated contact sheets: {output_paths}")
    return output_paths


def extract_audio(
    video_path: str,
    audio_path: Optional[str] = None,
    audio_bitrate: str = "128k",
    acodec: str = "mp3",
    max_duration: int = 300,  # 5 minutes
) -> str:
    if not audio_path:
        video_base_path, _ = os.path.splitext(video_path)
        audio_path = f"{video_base_path}.{acodec}"

    if os.path.exists(audio_path):
        return audio_path

    probe = ffmpeg.probe(video_path)
    video_duration = float(probe["streams"][0]["duration"])

    end_time_seconds = min(video_duration, max_duration)

    try:
        logger.info("Extracting audio from video")
        ffmpeg_input = ffmpeg.input(video_path, to=end_time_seconds)
        ffmpeg_input.output(
            audio_path,
            audio_bitrate=audio_bitrate,
            acodec=acodec,
        ).run(overwrite_output=True)
    except ffmpeg.Error as e:
        raise e
    return audio_path


def transcribe_audio(video_path: str, video_path_trimmed: str, start_time: str = "", end_time: str = "") -> str:
    if start_time:
        # Check if a VTT file already exists for the trimmed video
        dir_trimmed = os.path.dirname(video_path_trimmed)
        video_id_trimmed, _ = os.path.splitext(os.path.basename(video_path_trimmed))
        vtt_file = find_file(dir_trimmed, video_id_trimmed, (".vtt",), strict=True)
        if vtt_file:
            ensure_vtt_format(vtt_file)
            return vtt_file

        audio_path = find_file(dir_trimmed, video_id_trimmed, (".mp3",), strict=True)
    else:
        # Check if a VTT file already exists
        dir = os.path.dirname(video_path)
        video_id, _ = os.path.splitext(os.path.basename(video_path))
        vtt_file = find_file(dir, video_id, (".vtt",), strict=True)
        if vtt_file:
            ensure_vtt_format(vtt_file)
            return vtt_file

        audio_path = find_file(dir, video_id, (".mp3",), strict=True)

    if audio_path is None:
        audio_path = extract_audio(video_path_trimmed)

    vtt_path = audio_path.replace(".mp3", ".vtt")
    model = whisper.load_model("turbo")
    result = model.transcribe(audio_path)
    output_directory = os.path.dirname(audio_path)

    vtt_writer = get_writer("vtt", output_directory)
    vtt_writer(result, audio_path)
    vtt_path = increase_vtt_timestamps(vtt_path, offset_time=start_time)
    return vtt_path


def increase_vtt_timestamps(vtt_path: str, offset_time: str) -> str:
    offset_seconds = time_to_seconds(offset_time) if offset_time else 0
    if offset_seconds > 0:
        vtt = webvtt.read(vtt_path)
        for caption in vtt:
            caption.start = seconds_to_time(time_to_seconds(caption.start) + offset_seconds)
            caption.end = seconds_to_time(time_to_seconds(caption.end) + offset_seconds)
        vtt.save(vtt_path)
    return vtt_path


def extract_text_from_vtt(vtt_path: str, start_time: str = "", end_time: str = "") -> str:
    start_time = ensure_milliseconds(start_time)
    end_time = ensure_milliseconds(end_time)
    text = []

    vtt = webvtt.read(vtt_path)

    for caption in vtt.iter_slice(start=start_time, end=end_time):
        text.append(f"{str(caption.start)}: {str(caption.text)}")
    return "\n".join(text)


def ensure_vtt_format(vtt_path: str) -> None:
    with open(vtt_path, "r", encoding="utf-8") as file:
        lines = file.readlines()

    with open(vtt_path, "w", encoding="utf-8") as file:
        for line in lines:
            if line.strip() == "" and line.startswith(" "):  # Do not write line that contains only space
                continue
            else:
                file.write(line)


def ensure_milliseconds(time: str) -> None:
    """Ensure that the time string has milliseconds HH:MM:SS.mmm"""
    if not time:
        return time
    # validate time format
    if not re.match(r"^\d{2}:\d{2}:\d{2}(\.\d{3})?$", time):
        raise ValueError("Invalid time format:", time)
    # add .mmm if not included
    if time[-4] != ".":
        time += ".000"
    return time


def seconds_to_time(seconds: float) -> str:
    """Convert seconds SS to time string HH:MM:SS.mmm"""
    if type(seconds) is not float:
        try:
            seconds = float(seconds)
        except ValueError:
            raise ValueError("Invalid seconds format")
    hours = int(seconds // 3600)
    seconds %= 3600
    minutes = int(seconds // 60)
    seconds %= 60
    milliseconds = int((seconds % 1) * 1000)
    return f"{hours:02d}:{minutes:02d}:{int(seconds):02d}.{milliseconds:03d}"


def time_to_seconds(time_str: str) -> float:
    """Convert a time string HH:MM:SS(.mmm) and MM:SS().mmm) to seconds SS"""
    parts = list(map(float, re.split("[:.]", time_str)))
    if len(parts) == 1:
        return parts[0]
    elif len(parts) == 2:
        return int(parts[0]) * 60 + parts[1]
    elif len(parts) == 3:
        return int(parts[0]) * 3600 + int(parts[1]) * 60 + parts[2]
    elif len(parts) == 4:
        return int(parts[0]) * 3600 + int(parts[1]) * 60 + parts[2] + parts[3] / 1000
    else:
        raise ValueError("Invalid time format")


class YTDLogger(object):
    def debug(self, msg: str) -> None:
        # For compatibility with youtube-dl, both debug and info are passed into debug
        # You can distinguish them by the prefix '[debug] '
        if msg.startswith("[debug] "):
            logger.debug(msg)
        else:
            self.info(msg)

    def info(self, msg: str) -> None:
        logger.info(msg)
        pass

    def warning(self, msg: str) -> None:
        logger.warning(msg)
        pass

    def error(self, msg: str) -> None:
        logger.error(msg)


def ytd_progress_hook(d: dict) -> None:
    if d["status"] == "finished":
        logger.info("Done downloading, now converting ...")


class VideoReader(Tool):
    """
    Opens video from a youtube URL.
    Can access the video content, thumbnail, subtitles and audio.
    """

    action: type[Action] = WatchVideoAction
    observation: type[Observation] = VideoObservation
    cached: bool = True
    exp_path: str

    def execute_action(self, action: WatchVideoAction) -> VideoObservation:
        attachment_dir = os.path.join(self.exp_path, ATTACHMENT_DEFAULT_DIR)
        os.makedirs(attachment_dir, exist_ok=True)
        return get_video_observation(action.video_url, attachment_dir, action.start_time, action.end_time)



import sys

from tapeagents.llms import LLM, TrainableLLM
from tapeagents.renderers.pretty import PrettyRenderer
from tapeagents.studio import Studio
from tapeagents.team import TeamAgent, TeamTape


def try_chat(llm: LLM, studio: bool):
    # equilavent of https://microsoft.github.io/autogen/docs/tutorial/introduction
    comedy_duo = TeamAgent.create_initiator(
        name="Joe",
        llm=llm,
        system_prompt="Your name is Joe and you are a part of a duo of comedians.",
        teammate=TeamAgent.create(
            name="Cathy", llm=llm, system_prompt="Your name is Cathy and you are a part of a duo of comedians."
        ),
        max_calls=3,
        init_message="Hey Cathy, tell me a joke",
    )
    if studio:
        Studio(comedy_duo, TeamTape(context=None, steps=[]), PrettyRenderer()).launch()
    else:
        for event in comedy_duo.run(TeamTape(context=None, steps=[])):
            print(event.model_dump_json(indent=2))


if __name__ == "__main__":
    llm = TrainableLLM(
        base_url="https://api.together.xyz",
        model_name="meta-llama/Meta-Llama-3-70B-Instruct-Turbo",
        parameters=dict(temperature=0.7, max_tokens=512),
    )
    if len(sys.argv) == 2:
        if sys.argv[1] == "studio":
            try_chat(llm, studio=True)
        else:
            raise ValueError()
    elif len(sys.argv) == 1:
        try_chat(llm, studio=False)
    else:
        raise ValueError()



import time

from litellm import ChatCompletionMessageToolCall

from tapeagents.core import LLMOutput, Prompt, TrainingText
from tapeagents.llms.base import LLM, LLMEvent, LLMStream


class MockLLM(LLM):
    """A mock LLM implementation for testing purposes.

    This class simulates an LLM by returning predefined responses in a cyclic manner.
    It tracks the prompts it receives and maintains a call counter.

    Attributes:
        model_name (str): Name of the mock model, defaults to "mock"
        call_number (int): Counter for number of calls made to generate, defaults to 0
        mock_outputs (list[str]): List of predefined responses to cycle through
        prompts (list[Prompt]): List of received prompts
    """

    model_name: str = "mock"
    call_number: int = 0
    mock_outputs: list[str] = [
        "Agent: I'm good, thank you",
        "Agent: Sure, I worked at ServiceNow for 10 years",
        "Agent: I have 10 zillion parameters",
    ]
    mock_tool_calls: list[ChatCompletionMessageToolCall] = []
    prompts: list[Prompt] = []

    def generate(self, prompt: Prompt) -> LLMStream:
        def _implementation():
            self.prompts.append(prompt)
            output = self.mock_outputs[self.call_number % len(self.mock_outputs)]
            tool_call = (
                self.mock_tool_calls[self.call_number % len(self.mock_tool_calls)] if self.mock_tool_calls else None
            )
            time.sleep(0.01)
            yield LLMEvent(output=LLMOutput(content=output, tool_calls=[tool_call]))
            self.call_number += 1

        return LLMStream(_implementation(), prompt=prompt)

    def count_tokens(self, messages: list[dict] | str) -> int:
        return 42

    def make_training_text(self, prompt: Prompt, output: LLMOutput) -> TrainingText:
        return TrainingText(text="mock trace", n_predicted=10)



import json
from pprint import pprint

from litellm.types.utils import ChatCompletionMessageToolCall
from pydantic import TypeAdapter

from tapeagents.agent import Agent
from tapeagents.core import Action, Prompt
from tapeagents.dialog_tape import (
    AssistantStep,
    DialogContext,
    DialogTape,
    UserStep,
)
from tapeagents.environment import (
    Environment,
    ExternalObservationNeeded,
)
from tapeagents.llms import LiteLLM, LLMStream
from tapeagents.orchestrator import main_loop
from tapeagents.prompting import tape_to_messages
from tapeagents.tool_calling import ToolCalls, ToolResult, ToolSpec

MOCK_TOOL_ENV_PROMPT_TEMPLATE = """You will generate result of the following function call

{function_call}

You will output JSON of the following structure:
{{
    "result": ...
}}

You will output just the JSON and nothing else. Go!
"""


class MockToolEnvironment(Environment):
    def __init__(self, llm: LiteLLM):
        self.llm = llm

    def react(self, tape: DialogTape) -> DialogTape:
        # TODO: move prompting to a separate agent?
        action = tape.steps[-1]
        assert isinstance(action, Action)
        if isinstance(action, ToolCalls):
            for tc in action.tool_calls:
                prompt_text = MOCK_TOOL_ENV_PROMPT_TEMPLATE.format(function_call=tc.function)
                messages = [{"role": "user", "content": prompt_text}]
                for event in self.llm.generate(Prompt(messages=messages)):
                    completion = event.output
                    if completion and not isinstance(completion, str):
                        completion = completion.content
                    if completion:
                        result_json = json.loads(completion)
                        if "result" not in result_json:
                            raise ValueError("Result JSON should have 'result' key")
                        observation = ToolResult(content=json.dumps(result_json["result"]), tool_call_id=tc.id)
                        tape = tape.append(observation)
        elif isinstance(action, AssistantStep):
            self.raise_external_observation_needed(action)
        else:
            self.raise_unexpected_action(action)
        return tape


class FunctionCallingAgent(Agent[DialogTape]):
    def make_prompt(self, tape: DialogTape):
        assert tape.context
        return Prompt(tools=[t.model_dump() for t in tape.context.tools], messages=tape_to_messages(tape))

    def generate_steps(self, _, llm_stream: LLMStream):
        o = llm_stream.get_output()
        if o.content:
            yield AssistantStep(content=o.content)
        elif o.tool_calls:
            assert all(isinstance(tc, ChatCompletionMessageToolCall) for tc in o.tool_calls)
            yield ToolCalls.from_llm_output(o)
        else:
            raise ValueError(f"don't know what to do with message {o}")


TOOL_SCHEMAS = TypeAdapter(list[ToolSpec]).validate_python(
    [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather in a given location",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {"type": "string", "description": "The city and state, e.g. San Francisco, CA"},
                        "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                    },
                    "required": ["location"],
                },
            },
        }
    ]
)


def try_openai_function_calling():
    llm = LiteLLM(model_name="gpt-3.5-turbo")
    agent = FunctionCallingAgent.create(llm)
    tape = DialogTape(context=DialogContext(tools=TOOL_SCHEMAS), steps=[])

    for event in agent.run(tape.append(UserStep(content="What's the weather like in San Francisco, Tokyo"))):
        if event.step:
            print(event.step)

    assert event.final_tape
    tape = event.final_tape
    tool_call_step = tape.steps[-1]
    assert isinstance(tool_call_step, ToolCalls)
    tool_calls = tool_call_step.tool_calls

    tape = tape.append(
        ToolResult(
            tool_call_id=tool_calls[0].id,
            content="Cloudy, 13C",
        )
    ).append(
        ToolResult(
            tool_call_id=tool_calls[1].id,
            content="Sunny, 23C",
        )
    )

    for event in agent.run(tape):
        if event.step:
            print(event.step)


def try_openai_function_callling_with_environment():
    llm = LiteLLM(model_name="gpt-3.5-turbo")
    agent = FunctionCallingAgent.create(llm)
    tape = DialogTape(
        context=DialogContext(tools=TOOL_SCHEMAS),
        steps=[UserStep(content="What's the weather like in San Francisco, Tokyo")],
    )
    environment = MockToolEnvironment(llm)
    for s in tape.steps:
        print("USER STEP")
        pprint(s.model_dump(exclude_none=True))
    try:
        for event in main_loop(agent, tape, environment, max_loops=3):
            if ae := event.agent_event:
                if ae.step:
                    print("AGENT STEP")
                    pprint(ae.step.model_dump(exclude_none=True))
            elif event.observation:
                print("OBSERVATION")
                pprint(event.observation.model_dump(exclude_none=True))
    except ExternalObservationNeeded as e:
        assert isinstance(e.action, AssistantStep)
    print("Stopping, next user message is needed")


if __name__ == "__main__":
    try_openai_function_calling()
    try_openai_function_callling_with_environment()



import logging
from typing import Literal

from pydantic import BaseModel, Field, JsonValue
from typing_extensions import Self

from tapeagents.core import LLMOutputParsingFailureAction, SetNextNode, StopStep
from tapeagents.dialog_tape import AssistantStep, UserStep

from .error import (
    FormFillerStateError,
    InvalidFunctionSchemaError,
    UnknownFunctionError,
    UnknownFunctionParameterError,
    UnknownFunctionSchemaError,
    ValidFunctionParameterSkipError,
    ValidFunctionParameterValueError,
)
from .schema import FunctionSchema
from .steps import (
    AnswerFromFunctionSchema,
    CallFunction,
    FormFillerStep,
    FunctionCandidates,
    FunctionResult,
    GatherValuesThought,
    InspectFunction,
    NoAnswerFromFunctionSchema,
    RefuseInexistentFunction,
    RefuseInexistentFunctionParameter,
    RefuseInvalidFunctionParameterSkip,
    RefuseInvalidFunctionParameterValue,
    RefuseToEngage,
    RequestExitConfirmation,
    RequestFunction,
    RequestFunctionCallConfirmation,
    RequestFunctionParameters,
    ResolveFunction,
    UpdateFunctionParameters,
    VerifyValuesThought,
)
from .tape import FormFillerTape
from .types import FunctionName, ParameterName

logger = logging.getLogger(__name__)


class FormFillerState(BaseModel):
    """
    The state of the agent.
    Attributes:
        functions (dict[FunctionName, None]): The functions that are known to the agent (default: empty dict).
        function_schemas (dict[FunctionName, FunctionSchema]): The function schemas that are known to the agent
            (default: empty dict).
        function_parameters_filled (dict[FunctionName, dict[ParameterName, JsonValue]]): The function parameters
            that have been filled (default: empty dict).
        function_parameters_skipped (dict[FunctionName, list[ParameterName]]): The function parameters
            that have been skipped (default: empty dict).
        function_call_results (dict[FunctionName, JsonValue]): The results of the function calls
            that have been made (default: empty dict).
    """

    messages: list[UserStep | AssistantStep] = Field(
        default_factory=list, description="The messages that have been exchanged between the user and the assistant."
    )
    functions: dict[FunctionName, None] = Field(
        default_factory=dict, description="The functions that are known to the agent."
    )
    function_schemas: dict[FunctionName, FunctionSchema] = Field(
        default_factory=dict, description="The function schemas that are known to the agent."
    )
    function_parameters_filled: dict[FunctionName, dict[ParameterName, JsonValue]] = Field(
        default_factory=dict,
        alias="function_parameters",
        description="The function parameters that have been filled.",
    )
    function_parameters_skipped: dict[FunctionName, list[ParameterName]] = Field(
        default_factory=dict,
        description="The function parameters that have been skipped.",
    )
    raw_function_parameters: dict[FunctionName, dict[ParameterName, JsonValue]] = Field(
        default_factory=dict,
        alias="raw_function_parameters",
        description="The extracted function parameters that have not been verified yet.",
    )
    function_parameters_verified: dict[FunctionName, dict[ParameterName, JsonValue]] = Field(
        default_factory=dict,
        description="The function parameters that have been verified.",
    )
    function_call_results: dict[FunctionName, JsonValue] = Field(
        default_factory=dict,
        description="The results of the function calls that have been made.",
    )

    def add_function(self: Self, function: FunctionName) -> Self:
        return self.model_copy(update={"functions": self.functions | {function: None}})

    def add_function_schema(self: Self, function: FunctionName, schema: None | FunctionSchema) -> Self:
        """
        Adds a function schema to the state for the given function, overwriting any existing schema
        for it.
        The schema is assumed to be valid. No validation is performed.
        Args:
            function (FunctionName): The function name.
            schema (None | FunctionSchema): The function schema.
        """
        if isinstance(schema, FunctionSchema):
            return self.model_copy(update={"function_schemas": self.function_schemas | {function: schema}})
        elif schema is None:
            return self.model_copy()
        else:
            raise ValueError(f"Invalid schema: {schema}")

    def check_if_function_known(self: Self, function: FunctionName) -> Literal[True] | FormFillerStateError:
        """
        Checks that the function is known
        """
        if function in self.functions:
            return True
        return UnknownFunctionError(function=function, message=f"No function named {function} known.")

    def check_if_function_schema_known(self: Self, function: FunctionName) -> Literal[True] | FormFillerStateError:
        """
        Checks that the function and its schema are known
        """

        def _check() -> Literal[True] | FormFillerStateError:
            if function in self.function_schemas:
                return True
            return UnknownFunctionSchemaError(function=function, message=f"No function schema for function {function}")

        return self.check_if_function_known(function) and _check()

    def check_if_function_parameter_in_schema(
        self: Self, function: FunctionName, parameter: ParameterName
    ) -> Literal[True] | FormFillerStateError:
        """
        Checks that:
        - the function and its schema are known,
        - the parameter is in the schema
        """

        def _check() -> Literal[True] | FormFillerStateError:
            if (
                (parameters := self.function_schemas[function].parameters) is not None
                and parameters.properties is not None
                and parameter in parameters.properties
            ):
                return True
            return UnknownFunctionParameterError(function=function, parameter=parameter)

        return self.check_if_function_schema_known(function=function) and _check()

    def check_if_function_parameter_value_valid(
        self: Self,
        function: FunctionName,
        parameter: ParameterName,
        parameter_value: JsonValue,
    ) -> Literal[True] | FormFillerStateError:
        """
        Checks that:
        - the function and its schema are known,
        - the parameter is in the schema,
        - the parameter value is valid
        """

        def _validate() -> Literal[True] | FormFillerStateError:
            return self.function_schemas[function].validate_parameter_value(
                parameter=parameter,
                parameter_value=parameter_value,
            )

        return self.check_if_function_parameter_in_schema(function=function, parameter=parameter) and _validate()

    def check_if_function_parameter_is_skippable(
        self: Self,
        function: FunctionName,
        parameter: ParameterName,
    ) -> Literal[True] | FormFillerStateError:
        """
        Checks that:
        - the function and its schema are known,
        - the parameter is in the schema,
        - the parameter is skippable
        """

        def _check() -> Literal[True] | FormFillerStateError:
            if (function_schema := self.function_schemas.get(function)) is None:
                return UnknownFunctionSchemaError(
                    function=function,
                    message=f"No function schema for function {function}",
                )
            return function_schema.validate_parameter_skip(parameter=parameter)

        return self.check_if_function_parameter_in_schema(function=function, parameter=parameter) and _check()

    def check_if_function_parameters_valid(
        self: Self,
        function: FunctionName,
    ) -> Literal[True] | FormFillerStateError:
        """
        Checks that:
        - the function and its schema are known,
        - all filled parameter values are valid
        """

        def _validate() -> Literal[True] | FormFillerStateError:
            parameter_values = self.function_parameters_filled.get(function, {})
            return self.function_schemas[function].validate_parameter_values(parameter_values=parameter_values)

        return self.check_if_function_schema_known(function=function) and _validate()

    def check_if_function_schema_valid(
        self: Self,
        function: FunctionName,
        function_schema: FunctionSchema,
    ) -> Literal[True] | FormFillerStateError:
        """
        Checks that that the function schema is valid
        """
        if function_schema.name != function:
            return InvalidFunctionSchemaError(
                function=function,
                function_schema=function_schema,
                message=f"Function name in schema ({function_schema.name}) does not match function name ({function}).",
            )
        return function_schema.validate_parameters_schema()

    def check_if_function_return_value_valid(
        self: Self,
        function: FunctionName,
        return_value: JsonValue,
    ) -> Literal[True] | FormFillerStateError:
        """
        Checks that:
        - the function and its schema are known,
        - the return value is valid
        """
        if (function_schema := self.function_schemas.get(function)) is None:
            return UnknownFunctionSchemaError(
                function=function,
                message=f"No function schema for function {function}",
            )
        return self.check_if_function_schema_known(function=function) and function_schema.validate_return_value(
            return_value=return_value
        )

    def get_requestable_function_parameters(self: Self, function: FunctionName) -> list[ParameterName]:
        """
        Get the list of parameter names that can be requested for the given function.
        If the function is unknown, an empty list is returned.
        Args:
            function (FunctionName): The function name.
        Returns:
            list[ParameterName]: The list of parameters that can be requested for the given function.
        """
        if function not in self.function_schemas:
            return []
        return self.function_schemas[function].get_requestable_parameter_names(
            parameters_filled=self.function_parameters_filled.get(function, {}),
            parameters_skipped=self.function_parameters_skipped.get(function, []),
        )


def update_form_filler_state(state: FormFillerState, step: FormFillerStep) -> FormFillerState | FormFillerStateError:
    new_state = state.model_copy()
    match step:
        case UserStep() | AssistantStep():
            new_state.messages.append(step)
        case FunctionCandidates():
            for function in step.candidates:
                new_state = new_state.add_function(function.function)
        case InspectFunction():
            # check that the function is known
            if not (error := new_state.check_if_function_known(step.function)):
                return error.add_cause(step)
        case FunctionSchema():
            # check that the function is known
            if not (error := new_state.check_if_function_known(step.name)):
                return error.add_cause(step)
            # check that function schema is valid
            if not (error := new_state.check_if_function_schema_valid(step.name, step)):
                return error.add_cause(step)
            # add function schema to the state
            new_state = new_state.add_function_schema(step.name, step)
        case RequestFunction():
            # check that all functions are known
            if step.functions:
                errors = [
                    check
                    for check in (new_state.check_if_function_known(function) for function in step.functions)
                    if not check
                ]
                if errors:
                    assert isinstance(errors[0], FormFillerStateError)
                    return errors[0].add_cause(step)
        case CallFunction():
            # check that the function parameters are valid (also checks if the function and its schema are known)
            if not (error := new_state.check_if_function_parameters_valid(step.function)):
                return error.add_cause(step)
        case FunctionResult():
            # check that the function result is valid (also checks if the function and its schema are known)
            if not (error := new_state.check_if_function_return_value_valid(step.function, step.result)):
                return error.add_cause(step)
            # add function call result to the state
            new_state = new_state.model_copy(
                update={"function_call_results": new_state.function_call_results | {step.function: step.result}}
            )
        case RequestFunctionParameters():
            # check that the function parameters are in the schema (also checks if the function and its schema are known)
            errors = [
                check
                for check in (
                    new_state.check_if_function_parameter_in_schema(function=step.function, parameter=parameter)
                    for parameter in step.parameters
                )
                if not check
            ]
            if errors:
                assert isinstance(errors[0], FormFillerStateError)
                return errors[0].add_cause(step)
        case RequestFunctionCallConfirmation():
            # check that the function parameters are valid (also checks if the function and its schema are known)
            if not (error := new_state.check_if_function_parameters_valid(step.function)):
                return error.add_cause(step)
        case UpdateFunctionParameters():
            if step.assign:
                # check that each parameter value is valid (also checks if the function and its schema are known, and that the parameter is in the schema)
                errors = [
                    check
                    for check in (
                        new_state.check_if_function_parameter_value_valid(
                            function=step.function,
                            parameter=parameter,
                            parameter_value=parameter_value,
                        )
                        for parameter, parameter_value in step.assign.items()
                    )
                    if not check
                ]
                if errors:
                    assert isinstance(errors[0], FormFillerStateError)
                    return errors[0].add_cause(step)
                filled = new_state.function_parameters_filled
                filled[step.function] = filled.get(step.function, {}) | step.assign
            if step.skip:
                # check that each parameter is skippable (also checks if the function and its schema are known, and that the parameter is in the schema)
                errors = [
                    check
                    for check in (
                        new_state.check_if_function_parameter_is_skippable(function=step.function, parameter=parameter)
                        for parameter in step.skip
                    )
                    if not check
                ]
                if errors:
                    assert isinstance(errors[0], FormFillerStateError)
                    return errors[0].add_cause(step)
                filled = new_state.function_parameters_filled
                skipped = new_state.function_parameters_skipped
                if step.function not in skipped:
                    skipped[step.function] = []
                for k in step.skip:
                    filled.get(step.function, {}).pop(k, None)
                    skipped[step.function].append(k)
        case GatherValuesThought():
            # update raw function parameters
            new_state = new_state.model_copy(
                update={"raw_function_parameters": new_state.raw_function_parameters | {step.function: step.parameters}}
            )
        case VerifyValuesThought():
            # update function parameters verified
            new_state = new_state.model_copy(
                update={
                    "function_parameters_verified": new_state.function_parameters_verified
                    | {step.function: step.parameters}
                }
            )
        case RefuseInvalidFunctionParameterValue():
            # check that the parameter value is NOT valid (also checks if the function and its schema are known, and that the parameter is in the schema)
            if new_state.check_if_function_parameter_value_valid(
                function=step.function,
                parameter=step.parameter,
                parameter_value=step.parameter_value,
            ):
                return ValidFunctionParameterValueError(
                    function=step.function,
                    parameter=step.parameter,
                    parameter_value=step.parameter_value,
                ).add_cause(step)
        case RefuseInvalidFunctionParameterSkip():
            # check that the parameter is NOT skippable (also checks if the function and its schema are known, and that the parameter is in the schema)
            if new_state.check_if_function_parameter_is_skippable(
                function=step.function,
                parameter=step.parameter,
            ):
                return ValidFunctionParameterSkipError(
                    function=step.function,
                    parameter=step.parameter,
                ).add_cause(step)
        case RefuseInexistentFunctionParameter() | AnswerFromFunctionSchema() | NoAnswerFromFunctionSchema():
            # in all these cases, we just need to check that the function and its schema are known
            if not (error := new_state.check_if_function_schema_known(step.function)):
                return error.add_cause(step)
        case (
            ResolveFunction()
            | RequestExitConfirmation()
            | RefuseInexistentFunction()
            | RefuseToEngage()
            | LLMOutputParsingFailureAction()
            | SetNextNode()
            | StopStep()
        ):
            # in all these cases, there is nothing to check and nothing to update the state
            pass
        case _:
            # unknown step type: we should verify that they do not need to be handled
            logger.warning(
                f"Unknown step type: {step}. Verify that it does not need to be handled while updating form filler state. Will be ignored."
            )

    return new_state


def compute_form_filler_state(tape: FormFillerTape) -> FormFillerState | FormFillerStateError:
    # TODO: cache
    state = FormFillerState()
    for step in tape:
        result = update_form_filler_state(state, step)
        if isinstance(result, FormFillerStateError):
            return result
        state = result
    return state



import json
from typing import Literal

from tapeagents.core import Action, Tape, TapeMetadata


class DummyStep(Action):
    pass


TestTape = Tape[None, DummyStep]


def test_add_two_tapes():
    tape1 = TestTape(steps=[DummyStep(), DummyStep()])
    tape2 = TestTape(steps=[DummyStep()])

    result = tape1 + tape2

    assert len(result.steps) == 3
    assert isinstance(result, Tape)
    assert result.metadata.n_added_steps == 1


def test_add_tape_with_iterable():
    tape = TestTape(steps=[DummyStep()])
    steps = [DummyStep(), DummyStep()]

    result = tape + steps

    assert len(result.steps) == 3
    assert isinstance(result, Tape)
    assert result.metadata.n_added_steps == 2


def test_add_empty_tape():
    tape1 = TestTape(steps=[DummyStep()])
    tape2 = TestTape(steps=[])

    result = tape1 + tape2

    assert len(result.steps) == 1
    assert isinstance(result, Tape)
    assert result.metadata.n_added_steps == 0


def test_add_to_empty_tape():
    tape = TestTape(steps=[])
    steps = [DummyStep(), DummyStep()]

    result = tape + steps

    assert len(result.steps) == 2
    assert isinstance(result, Tape)
    assert result.metadata.n_added_steps == 2


def test_getitem_single_step():
    tape = TestTape(steps=[DummyStep(), DummyStep(), DummyStep()])

    step = tape[1]

    assert isinstance(step, DummyStep)
    assert step == tape.steps[1]


def test_getitem_slice():
    tape = TestTape(steps=[DummyStep(), DummyStep(), DummyStep()], metadata=TapeMetadata(author="test"))

    sliced_tape = tape[1:3]

    assert isinstance(sliced_tape, Tape)
    assert len(sliced_tape.steps) == 2
    assert sliced_tape.steps[0] == tape.steps[1]
    assert sliced_tape.steps[1] == tape.steps[2]
    assert sliced_tape.metadata.author is None


def test_getitem_slice_empty():
    tape = TestTape(steps=[DummyStep(), DummyStep(), DummyStep()])

    sliced_tape = tape[3:3]

    assert isinstance(sliced_tape, Tape)
    assert len(sliced_tape.steps) == 0
    assert sliced_tape.metadata.author is None


def test_append_single_step():
    tape = TestTape(steps=[DummyStep()])
    new_step = DummyStep()

    result = tape.append(new_step)

    assert len(result.steps) == 2
    assert result.steps[-1] == new_step
    assert isinstance(result, Tape)
    # check that metadata is the same as a brand new one
    assert result.metadata == TapeMetadata(id=result.metadata.id)


def test_append_multiple_steps():
    tape = TestTape(steps=[DummyStep()])
    new_step1 = DummyStep()
    new_step2 = DummyStep()

    result = tape.append(new_step1).append(new_step2)

    assert len(result.steps) == 3
    assert result.steps[-2] == new_step1
    assert result.steps[-1] == new_step2
    assert isinstance(result, Tape)
    # check that metadata is the same as a brand new one
    assert result.metadata == TapeMetadata(id=result.metadata.id)


def test_append_to_empty_tape():
    tape = TestTape(steps=[])
    new_step = DummyStep()

    result = tape.append(new_step)

    assert len(result.steps) == 1
    assert result.steps[0] == new_step
    assert isinstance(result, Tape)
    # check that metadata is the same as a brand new one
    assert result.metadata == TapeMetadata(id=result.metadata.id)


def test_with_new_id():
    tape = TestTape(steps=[DummyStep(), DummyStep()], metadata=TapeMetadata(author="test"))

    new_tape = tape.with_new_id()

    assert isinstance(new_tape, Tape)
    assert new_tape.metadata.id != tape.metadata.id
    assert new_tape.metadata.author is None
    assert len(new_tape.steps) == len(tape.steps)
    assert new_tape.steps == tape.steps


def test_llm_dict():
    class TestStep(Action):
        kind: Literal["test_step"] = "test_step"
        data: str = "test_data"

    step = TestStep()
    llm_dict = step.llm_dict()

    assert "metadata" not in llm_dict
    assert llm_dict["kind"] == "test_step"
    assert llm_dict["data"] == "test_data"


def test_llm_dict_excludes_none_values():
    class TestStep(Action):
        kind: Literal["test_step"] = "test_step"
        data: str | None = None

    step = TestStep()
    llm_dict = step.llm_dict()

    assert "data" not in llm_dict
    assert llm_dict["kind"] == "test_step"


def test_llm_view():
    class TestStep(Action):
        kind: Literal["test_step"] = "test_step"
        data: str = "test_data"

    step = TestStep()
    llm_view = step.llm_view()

    expected_output = json.dumps(step.llm_dict(), indent=2, ensure_ascii=False)
    assert llm_view == expected_output


def test_llm_view_no_indent():
    class TestStep(Action):
        kind: Literal["test_step"] = "test_step"
        data: str = "test_data"

    step = TestStep()
    llm_view = step.llm_view(indent=None)

    expected_output = json.dumps(step.llm_dict(), indent=None, ensure_ascii=False)
    assert llm_view == expected_output


def test_llm_view_excludes_none_values():
    class TestStep(Action):
        kind: Literal["test_step"] = "test_step"
        data: str | None = None

    step = TestStep()
    llm_view = step.llm_view()

    expected_output = json.dumps(step.llm_dict(), indent=2, ensure_ascii=False)
    assert llm_view == expected_output



from typing import Literal, Optional

from pydantic import Field

from tapeagents.core import Action, Observation
from tapeagents.tools.base import Tool
from tapeagents.tools.converters import (
    FileConversionException,
    FileConverter,
    FileConverterOptions,
    UnsupportedFormatException,
)


def read_document(path: str, file_converter_options: Optional[FileConverterOptions] = None) -> tuple[str, str | None]:
    """Read a document, file or image and and convert it to Markdown."""
    try:
        text = ""
        error = None
        text = FileConverter(file_converter_options=file_converter_options).convert(path).text_content
    except UnsupportedFormatException as e:
        error = f"Failed to read document {path}: {e}"
    except FileConversionException as e:
        error = f"Failed to read document {path}: {e}"
    except Exception as e:
        error = f"Failed to read document {path}: {e}"
    return text, error


class DocumentObservation(Observation):
    kind: Literal["document_observation"] = "document_observation"
    text: str
    error: str | None = None


class ReadLocalDocumentAction(Action):
    """
    Action that loads the document, file or image and converts it to Markdown.
    """

    kind: Literal["read_local_document_action"] = "read_local_document_action"
    path: str = Field(description="path of the document")


class DocumentReader(Tool):
    """
    Tool to read a document and convert it to Markdown.
    """

    action: type[Action] = ReadLocalDocumentAction
    observation: type[Observation] = DocumentObservation
    file_converter_options: Optional[FileConverterOptions] = None

    def execute_action(self, action: ReadLocalDocumentAction) -> DocumentObservation:
        text, error = read_document(action.path, self.file_converter_options)
        return DocumentObservation(text=text, error=error)
